from typing import Optional


class RuptureProduct(object):
    def __init__(
        self,
        product_code,           # type: str
        product_name,           # type: str
        products_composition,   # type: str
        product_quantity=0      # type: Optional[int]
    ):
        # type: (...) -> None

        self.product_code = product_code
        self.product_name = product_name
        self.products_composition = products_composition
        self.product_quantity = product_quantity



import json

from xml.etree import cElementTree as eTree

from ._Kds import Kds
from ._PickupArea import PickupArea
from ._KdsView import KdsView
from ._ZoomLevel import ZoomLevel


class KdsModel(object):
    def __init__(self, kds, kds_id):
        # type: (Kds, int) -> None
        self.kds = kds
        self.kds_id = kds_id

    def get_model_xml(self):
        kds_model_xml = eTree.Element("KdsModel", kdsId=str(self.kds_id))
        eTree.SubElement(kds_model_xml, "Language", name=self.kds.language)
        eTree.SubElement(kds_model_xml, "Title", name=self.kds.title)

        self.create_brands_xml(kds_model_xml)
        self.create_bump_bar_xml(kds_model_xml)
        self.create_modifiers_xml(kds_model_xml)
        self.create_pickup_area_xml(kds_model_xml=kds_model_xml)
        self.create_views_xml(kds_model_xml)

        return kds_model_xml

    def create_brands_xml(
        self,
        kds_model_xml  # type: eTree.Element
    ):
        # type: (...) -> None

        brands_xml = eTree.SubElement(
            kds_model_xml,
            "Brands"
        )
        if self.kds.brands:
            for brand in self.kds.brands:
                eTree.SubElement(
                    brands_xml,
                    "Brand",
                    name=brand["brand"],
                    color=brand["color"]
                )

    def create_bump_bar_xml(self, kds_model_xml):
        bump_bar_xml = eTree.SubElement(kds_model_xml, "BumpBar")

        if self.kds.bump_bar:
            self.create_bump_bar_command_xml(self.kds.bump_bar, bump_bar_xml)

    @staticmethod
    def create_bump_bar_command_xml(bump_bar, bump_bar_xml):
        name = bump_bar.name
        simulate = str(bump_bar.simulate_bump_bar)

        bump_bar_xml = eTree.SubElement(bump_bar_xml, "BumpBar", name=name, simulate=simulate)
        for command in bump_bar.commands:
            eTree.SubElement(bump_bar_xml, "Command", name=command.name, value=str.join("|", command.keys))

    def create_modifiers_xml(self, kds_model_xml):
        modifiers_xml = eTree.SubElement(kds_model_xml, "ModifierPrefixes")
        for modifier in self.kds.modifiers:
            eTree.SubElement(modifiers_xml, "Prefix", name=modifier.name, value=modifier.description)

    def create_pickup_area_xml(
        self,           # type: KdsModel
        kds_model_xml,  # type: eTree.Element
    ):
        # type (...) -> None

        pickup_areas_xml = eTree.SubElement(kds_model_xml, "PickupAreas")
        if not self.kds.pickup_config:
            return

        for area in self.kds.pickup_config.pickup_areas:
            eTree.SubElement(
                pickup_areas_xml,
                "PickupArea",
                name=area.pickup_name,
                sectors=self._get_sectors_json(area)
            )

    @staticmethod
    def _get_sectors_json(
        area,       # type: PickupArea
    ):
        # type (...) -> None

        sectors = []
        for sector in area.sectors:
            sectors.append({
                "name": sector.name,
                "excludedPos": sector.excluded_pos,
            })

        return json.dumps(sectors)

    def create_views_xml(self, kds_model_xml):
        views_xml = eTree.SubElement(kds_model_xml, "Views")
        for view in self.kds.views:
            zoom = view.zoom_levels[view.selected_zoom]
            selected_view = str(view.selected).lower()

            view_xml = eTree.SubElement(views_xml, "View", name=view.name, selected=selected_view)
            eTree.SubElement(view_xml, "Selected", state=str(view.selected))
            eTree.SubElement(view_xml, "ProductionView", value=view.production_view)
            eTree.SubElement(view_xml, "AskPickupArea", value=view.ask_pickup_area)
            eTree.SubElement(view_xml, "DisplayMode", value=view.display_mode)
            eTree.SubElement(view_xml, "DisplayType", value=view.display_type)
            self._create_layout(view_xml, view, zoom)
            eTree.SubElement(view_xml, "Title", value=view.title)
            self._create_cell_format(view, view_xml)
            eTree.SubElement(view_xml, "Carousel", state=str(view.carousel))
            eTree.SubElement(view_xml, "ShowItems", state=str(view.show_items))
            eTree.SubElement(view_xml, "SelectedOrderId", value=str(view.selected_order_id))
            eTree.SubElement(view_xml, "SelectedZoom", value=str(view.selected_zoom))

            settings_xml = eTree.SubElement(view_xml, "Settings")
            eTree.SubElement(settings_xml, "Setting", name="cellDoubleClick", value=view.on_double_click)
            eTree.SubElement(settings_xml, "Setting", name="onSellClick", value=view.on_sell_click)

            thresholds_xml = eTree.SubElement(view_xml, "Thresholds")
            for threshold in view.thresholds:
                thresh_hold_percentage = (str(threshold.percentage) if threshold.percentage else "")
                value = str(threshold.time) + threshold.color + thresh_hold_percentage
                eTree.SubElement(thresholds_xml, "Threshold", value=value)

            statistics_xml = eTree.SubElement(view_xml, "Statistics")
            for statistic in view.statistics:
                eTree.SubElement(statistics_xml, "Statistic", name=statistic.name, format=statistic.format)

            if view.auto_order_command:
                time = str(view.auto_order_command.time)
                command = view.auto_order_command.command
                eTree.SubElement(view_xml, "AutoOrderCommand", time=time, command=command)

            eTree.SubElement(view_xml, "ColoredArea", area=view.colored_area.lower())
            eTree.SubElement(view_xml, "ViewWithActions", value=str(view.view_with_actions).lower())

    @staticmethod
    def _create_cell_format(view, view_xml):
        header_macro = view.header_text_macro
        footer_macro = view.footer_text_macro
        eTree.SubElement(view_xml, "CellFormat", header=header_macro, footer=footer_macro, body=view.body_text_macro,
                         saleLine=view.body_sale_line_macro)

    @staticmethod
    def _create_layout(
        view_xml,       # type: eTree.Element
        view,           # type: KdsView
        zoom,           # type: ZoomLevel
    ):
        # type: (...) -> None

        eTree.SubElement(
            view_xml,
            "Layout",
            rows=zoom.rows,
            lines=zoom.lines,
            cols=zoom.cols,
            fontsize=view.custom_font_size or zoom.fontsize,
            fontcolor=view.custom_font_color or ''
        )



import os

from application.model.configuration import SystemPaths
from application.util import LoggerUtil
from _CommonService import CommonService

logger = LoggerUtil.get_logger_name(__name__)


class StaticConfigurationService(object):
    @staticmethod
    def update_static_databases():
        # type: () -> None

        try:
            logger.info("Updating static databases")
            static_databases = os.path.join(SystemPaths.static_configuration_folder.value, "databases")
            if not os.path.exists(static_databases):
                logger.info("No static database to update")
                return

            CommonService.copy_all_content(static_databases, SystemPaths.genesis_databases_path.value)
            logger.info("Static databases updated")
        except Exception as _:
            logger.exception("Error updating static databases")

    @staticmethod
    def update_static_loaders():
        # type: () -> None

        try:
            logger.info("Updating static loaders")
            static_loaders = os.path.join(SystemPaths.static_configuration_folder.value, "loaders")
            if not os.path.exists(static_loaders):
                logger.info("No static loader to update")
                return

            for directory in os.listdir(static_loaders):
                static_directory_path = os.path.join(static_loaders, directory)
                directory_path = os.path.join(SystemPaths.genesis_data_path.value, directory)
                CommonService.create_dir(directory_path, False)
                CommonService.copy_all_content(static_directory_path, directory_path)
        except Exception as _:
            logger.exception("Error updating static loaders")

    @staticmethod
    def run_databases_script():
        # type: () -> None

        try:
            logger.info("Running static scripts")
            static_scripts = os.path.join(SystemPaths.static_configuration_folder.value, "scripts")
            if not os.path.exists(static_scripts):
                logger.info("No static scripts to update")
                return

            for script in os.listdir(static_scripts):
                if not script.endswith(".sql"):
                    continue

                database_name = script.replace(".sql", ".db")
                logger.info("Running script for {}".format(database_name))
                script_path = os.path.join(static_scripts, script)
                database = os.path.join(SystemPaths.genesis_databases_path.value, database_name)
                if not os.path.exists(database):
                    logger.info("No database found {}".format(database))
                    continue

                backup_database = database.replace(".db", "_bkp.db")
                try:
                    CommonService.duplicate_file(database, backup_database)
                    CommonService.execute_scripts_commands(database, script_path, logger)
                    logger.info("Script executed for {}".format(database))
                except Exception as _:
                    logger.exception("Error running script for {}".format(script))
                    CommonService.remove_file(database)
                    CommonService.duplicate_file(backup_database, database)
                finally:
                    CommonService.remove_file(backup_database)
        except Exception as _:
            logger.exception("Error running static scripts")



from application.domain import Order, CustomOrderProperty
from typing import List  # noqa
from xml.etree import cElementTree as eTree


class TablePictureToOrdersParser(object):
    @staticmethod
    def parse(table_picture):
        # type: (str) -> List[Order]
        xml_table_picture = eTree.XML(table_picture)
        xml_orders = xml_table_picture.findall("Orders/Order")
        result_orders = []
        for order in xml_orders:
            order_custom_property_list = []
            for order_custom_property in order.findall("CustomOrderProperties/OrderProperty"):
                order_custom_property_list.append(
                    CustomOrderProperty(order_custom_property.get("key"), order_custom_property.get("value")))

            result_orders.append(Order(
                order.get("orderId"),
                float(order.get("totalGross")) + float(order.get("discountAmount")),
                float(order.get("tip")),
                order.get("posId"),
                order_custom_property_list,
                float(order.get("dueAmount"))
            ))

        return result_orders



from application.service import LoaderUpdaterService


class LoaderUpdaterInteractor(object):
    def __init__(self, loader_updater_service):
        # type: (LoaderUpdaterService) -> None # noqa

        self.loader_updater_service = loader_updater_service

    def update_loaders(self):
        # type: () -> None

        return self.loader_updater_service.update_loaders(True)




data = (
'dwaen',    # 0x00
'dwaenj',    # 0x01
'dwaenh',    # 0x02
'dwaed',    # 0x03
'dwael',    # 0x04
'dwaelg',    # 0x05
'dwaelm',    # 0x06
'dwaelb',    # 0x07
'dwaels',    # 0x08
'dwaelt',    # 0x09
'dwaelp',    # 0x0a
'dwaelh',    # 0x0b
'dwaem',    # 0x0c
'dwaeb',    # 0x0d
'dwaebs',    # 0x0e
'dwaes',    # 0x0f
'dwaess',    # 0x10
'dwaeng',    # 0x11
'dwaej',    # 0x12
'dwaec',    # 0x13
'dwaek',    # 0x14
'dwaet',    # 0x15
'dwaep',    # 0x16
'dwaeh',    # 0x17
'doe',    # 0x18
'doeg',    # 0x19
'doegg',    # 0x1a
'doegs',    # 0x1b
'doen',    # 0x1c
'doenj',    # 0x1d
'doenh',    # 0x1e
'doed',    # 0x1f
'doel',    # 0x20
'doelg',    # 0x21
'doelm',    # 0x22
'doelb',    # 0x23
'doels',    # 0x24
'doelt',    # 0x25
'doelp',    # 0x26
'doelh',    # 0x27
'doem',    # 0x28
'doeb',    # 0x29
'doebs',    # 0x2a
'does',    # 0x2b
'doess',    # 0x2c
'doeng',    # 0x2d
'doej',    # 0x2e
'doec',    # 0x2f
'doek',    # 0x30
'doet',    # 0x31
'doep',    # 0x32
'doeh',    # 0x33
'dyo',    # 0x34
'dyog',    # 0x35
'dyogg',    # 0x36
'dyogs',    # 0x37
'dyon',    # 0x38
'dyonj',    # 0x39
'dyonh',    # 0x3a
'dyod',    # 0x3b
'dyol',    # 0x3c
'dyolg',    # 0x3d
'dyolm',    # 0x3e
'dyolb',    # 0x3f
'dyols',    # 0x40
'dyolt',    # 0x41
'dyolp',    # 0x42
'dyolh',    # 0x43
'dyom',    # 0x44
'dyob',    # 0x45
'dyobs',    # 0x46
'dyos',    # 0x47
'dyoss',    # 0x48
'dyong',    # 0x49
'dyoj',    # 0x4a
'dyoc',    # 0x4b
'dyok',    # 0x4c
'dyot',    # 0x4d
'dyop',    # 0x4e
'dyoh',    # 0x4f
'du',    # 0x50
'dug',    # 0x51
'dugg',    # 0x52
'dugs',    # 0x53
'dun',    # 0x54
'dunj',    # 0x55
'dunh',    # 0x56
'dud',    # 0x57
'dul',    # 0x58
'dulg',    # 0x59
'dulm',    # 0x5a
'dulb',    # 0x5b
'duls',    # 0x5c
'dult',    # 0x5d
'dulp',    # 0x5e
'dulh',    # 0x5f
'dum',    # 0x60
'dub',    # 0x61
'dubs',    # 0x62
'dus',    # 0x63
'duss',    # 0x64
'dung',    # 0x65
'duj',    # 0x66
'duc',    # 0x67
'duk',    # 0x68
'dut',    # 0x69
'dup',    # 0x6a
'duh',    # 0x6b
'dweo',    # 0x6c
'dweog',    # 0x6d
'dweogg',    # 0x6e
'dweogs',    # 0x6f
'dweon',    # 0x70
'dweonj',    # 0x71
'dweonh',    # 0x72
'dweod',    # 0x73
'dweol',    # 0x74
'dweolg',    # 0x75
'dweolm',    # 0x76
'dweolb',    # 0x77
'dweols',    # 0x78
'dweolt',    # 0x79
'dweolp',    # 0x7a
'dweolh',    # 0x7b
'dweom',    # 0x7c
'dweob',    # 0x7d
'dweobs',    # 0x7e
'dweos',    # 0x7f
'dweoss',    # 0x80
'dweong',    # 0x81
'dweoj',    # 0x82
'dweoc',    # 0x83
'dweok',    # 0x84
'dweot',    # 0x85
'dweop',    # 0x86
'dweoh',    # 0x87
'dwe',    # 0x88
'dweg',    # 0x89
'dwegg',    # 0x8a
'dwegs',    # 0x8b
'dwen',    # 0x8c
'dwenj',    # 0x8d
'dwenh',    # 0x8e
'dwed',    # 0x8f
'dwel',    # 0x90
'dwelg',    # 0x91
'dwelm',    # 0x92
'dwelb',    # 0x93
'dwels',    # 0x94
'dwelt',    # 0x95
'dwelp',    # 0x96
'dwelh',    # 0x97
'dwem',    # 0x98
'dweb',    # 0x99
'dwebs',    # 0x9a
'dwes',    # 0x9b
'dwess',    # 0x9c
'dweng',    # 0x9d
'dwej',    # 0x9e
'dwec',    # 0x9f
'dwek',    # 0xa0
'dwet',    # 0xa1
'dwep',    # 0xa2
'dweh',    # 0xa3
'dwi',    # 0xa4
'dwig',    # 0xa5
'dwigg',    # 0xa6
'dwigs',    # 0xa7
'dwin',    # 0xa8
'dwinj',    # 0xa9
'dwinh',    # 0xaa
'dwid',    # 0xab
'dwil',    # 0xac
'dwilg',    # 0xad
'dwilm',    # 0xae
'dwilb',    # 0xaf
'dwils',    # 0xb0
'dwilt',    # 0xb1
'dwilp',    # 0xb2
'dwilh',    # 0xb3
'dwim',    # 0xb4
'dwib',    # 0xb5
'dwibs',    # 0xb6
'dwis',    # 0xb7
'dwiss',    # 0xb8
'dwing',    # 0xb9
'dwij',    # 0xba
'dwic',    # 0xbb
'dwik',    # 0xbc
'dwit',    # 0xbd
'dwip',    # 0xbe
'dwih',    # 0xbf
'dyu',    # 0xc0
'dyug',    # 0xc1
'dyugg',    # 0xc2
'dyugs',    # 0xc3
'dyun',    # 0xc4
'dyunj',    # 0xc5
'dyunh',    # 0xc6
'dyud',    # 0xc7
'dyul',    # 0xc8
'dyulg',    # 0xc9
'dyulm',    # 0xca
'dyulb',    # 0xcb
'dyuls',    # 0xcc
'dyult',    # 0xcd
'dyulp',    # 0xce
'dyulh',    # 0xcf
'dyum',    # 0xd0
'dyub',    # 0xd1
'dyubs',    # 0xd2
'dyus',    # 0xd3
'dyuss',    # 0xd4
'dyung',    # 0xd5
'dyuj',    # 0xd6
'dyuc',    # 0xd7
'dyuk',    # 0xd8
'dyut',    # 0xd9
'dyup',    # 0xda
'dyuh',    # 0xdb
'deu',    # 0xdc
'deug',    # 0xdd
'deugg',    # 0xde
'deugs',    # 0xdf
'deun',    # 0xe0
'deunj',    # 0xe1
'deunh',    # 0xe2
'deud',    # 0xe3
'deul',    # 0xe4
'deulg',    # 0xe5
'deulm',    # 0xe6
'deulb',    # 0xe7
'deuls',    # 0xe8
'deult',    # 0xe9
'deulp',    # 0xea
'deulh',    # 0xeb
'deum',    # 0xec
'deub',    # 0xed
'deubs',    # 0xee
'deus',    # 0xef
'deuss',    # 0xf0
'deung',    # 0xf1
'deuj',    # 0xf2
'deuc',    # 0xf3
'deuk',    # 0xf4
'deut',    # 0xf5
'deup',    # 0xf6
'deuh',    # 0xf7
'dyi',    # 0xf8
'dyig',    # 0xf9
'dyigg',    # 0xfa
'dyigs',    # 0xfb
'dyin',    # 0xfc
'dyinj',    # 0xfd
'dyinh',    # 0xfe
'dyid',    # 0xff
)



from typing import Optional


class RuptureItem(object):
    def __init__(
        self,
        product_code,       # type: str
        product_name,       # type: str
        product_part_code,  # type: str
        product_quantity=0  # type: Optional[int]
    ):
        # type: (...) -> None

        self.product_code = product_code
        self.product_name = product_name
        self.product_part_code = product_part_code
        self.product_quantity = product_quantity



# ssn.py - functions for handling Swiss social security numbers
#
# Copyright (C) 2014 Denis Krienbuehl
# Copyright (C) 2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Swiss social security number ("Sozialversicherungsnummer").

Also known as "Neue AHV Nummer". The Swiss Sozialversicherungsnummer is used
to identify individuals for taxation and pension purposes.

The number is validated using EAN-13, though dashes are substituted for dots.

More information:

* https://en.wikipedia.org/wiki/National_identification_number#Switzerland
* https://de.wikipedia.org/wiki/Sozialversicherungsnummer#Versichertennummer

>>> validate('7569217076985')
'7569217076985'
>>> validate('756.9217.0769.85')
'7569217076985'
>>> validate('756.9217.0769.84')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('123.4567.8910.19')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> format('7569217076985')
'756.9217.0769.85'
"""

from stdnum import ean
from stdnum.exceptions import *
from stdnum.util import clean


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' .').strip()


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '.'.join((number[:3], number[3:7], number[7:11], number[11:]))


def validate(number):
    """Check if the number is a valid Swiss Sozialversicherungsnummer."""
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    if not number.startswith('756'):
        raise InvalidComponent()
    return ean.validate(number)


def is_valid(number):
    """Check if the number is a valid Swiss Sozialversicherungsnummer."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import json
import logging

from xml.etree import cElementTree as eTree

import sysactions

from actions.util import (
    get_authorization,
    get_user_level,
)
from model.customexception import (
    ErrorApplyingDiscount,
    DiscountMinLevelException,
    UnexpectedResponse,
)
from typing import Optional, Dict

from actions.sale_line.set_custom_properties import set_custom_properties
from bustoken import (
    TK_NCOUPON_APPLY_COUPON,
    TK_NCOUPON_APPLY_BENEFIT,
    create_token,
    MSGPRT_LOW,
)
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    MBTimeout,
    MBException,
)
from mw_helper import (
    show_keyboard,
)

from sysactions import action, get_model, get_current_order, show_messagebox, get_posot, translate_message

from actions import mb_context
from custom_sysactions import user_authorized_to_execute_action

from ._ApplyCouponErrorCode import ApplyCouponErrorCode
from ._CustomError import CustomError
from ._TranslatedMessage import TranslatedMessage

NCOUPON_MESSAGE_GROUP = "113"
TK_NCOUPON_ERROR = create_token(
    priority=MSGPRT_LOW,
    message_grourp=NCOUPON_MESSAGE_GROUP,
    number="F001",
)

logger = logging.getLogger("PosActions")


@action
@user_authorized_to_execute_action
def do_apply_coupons(pos_id, table_id=None, seat=''):
    coupon_code = show_keyboard(
        posid=pos_id,
        message="$ENTER_COUPON_CODE",
        defvalue="",
        title="",
        numpad=False,
        buttons='$OK|$PREVIOUS',
    )
    if coupon_code is None:
        return False

    if seat == '':
        try:
            apply_coupon(pos_id, table_id=table_id, coupon_code=coupon_code)
        except DiscountMinLevelException as ex:
            show_messagebox(
                posid=pos_id,
                message="$NDISCOUNT_ERROR_USER_MIN_LEVEL",
            )
            user_was_authorized, user_auth = get_authorization(
                pos_id=pos_id,
                min_level=ex.min_level_required,
                model=get_model(pos_id),
            )
            if user_was_authorized:
                user_xml_str = sysactions.get_user_information(user_number=user_auth)
                user_element = eTree.XML(user_xml_str).find("user")
                user_auth_level = int(user_element.attrib["Level"])

                apply_coupon(
                    pos_id=pos_id,
                    table_id=table_id,
                    coupon_code=coupon_code,
                    user_auth_level=user_auth_level,
                )
                return True
        return

    _apply_coupon_and_set_seat(coupon_code, pos_id, seat, table_id)


def apply_coupon(
    pos_id,                 # type: str
    table_id=None,          # type: Optional[str]
    coupon_code=None,       # type: Optional[str]
    user_auth_level=-1,     # type: Optional[int]
    benefit=None,           # type: Optional[Dict]
    applier=None,           # type: Optional[str]
):
    # type: (...) -> bool

    model = get_model(pos_id)
    order = get_current_order(model)
    message = None

    if user_auth_level == -1:
        user_auth_id = int(model.find("Operator").get("id"))
        user_auth_level = get_user_level(user_auth_id)

    try:
        originator_pos_id = int(order.get("originatorId")[3:]) if order else None
        if benefit is None:
            coupon_code = coupon_code.upper()
            data = json.dumps(dict(
                posId=pos_id,
                originatorPosId=originator_pos_id,
                couponId=coupon_code,
                user_auth_level=user_auth_level
            ))

            response = mb_context.MB_EasySendMessage("nCoupon", TK_NCOUPON_APPLY_COUPON, FM_PARAM, data)
        else:
            data = json.dumps(
                dict(
                    posId=pos_id,
                    originatorPosId=originator_pos_id,
                    benefit=benefit,
                    user_auth_level=user_auth_level,
                    applier=applier,
                )
            )

            response = mb_context.MB_EasySendMessage("nCoupon", TK_NCOUPON_APPLY_BENEFIT, FM_PARAM, data)

        if response.token == TK_SYS_ACK:
            if order and order.get('state') == "TOTALED":
                message = "$SUCCESS_COUPON_APPLIED"
            else:
                message = "$SUCCESS_NOT_TOTALED_COUPON_APPLIED"
            if table_id is not None:
                from table_actions import process_discount_in_table
                process_discount_in_table(pos_id, order, table_id)
            return True
        elif response.token == TK_NCOUPON_ERROR:
            error_code = response.data.split("\0")[0]
            if error_code == ApplyCouponErrorCode.COUPON_LOCKED:
                message = "$ERROR_COUPON_LOCKED_BY_ANOTHER_REQUEST"
            elif error_code == ApplyCouponErrorCode.COUPON_ALREADY_USED:
                message = "$ERROR_COUPON_APPLIED"
            elif error_code == ApplyCouponErrorCode.COUPON_NOT_FOUND:
                message = "$ERROR_COUPON_NOT_FOUND"
            elif error_code == ApplyCouponErrorCode.CUSTOM_ERROR:
                data_json = json.loads(response.data.split("\0")[1].decode("utf-8"))
                custom_error = _build_custom_error(data_json)
                message = _get_translated_message(
                    pos_id=pos_id,
                    custom_error=custom_error,
                )
            elif error_code == ApplyCouponErrorCode.NDISCOUNT_ERROR:
                ndiscount_error = response.data.split("\0")[1]
                ndiscount_error_code = ndiscount_error.split("\0")[0]
                if ndiscount_error_code == "NDISCOUNT_ERROR_USER_MIN_LEVEL":
                    raise DiscountMinLevelException(min_level_required=int(ndiscount_error_code.split("\0")[1]))

                raise ErrorApplyingDiscount(error_code=ndiscount_error_code)
            else:
                logger.error("Unknown error code from nCoupon: {}".format(error_code))
                message = "$ERROR_COUPON_APPLIED"
        else:
            raise UnexpectedResponse("nCoupon", response)

    except (MBTimeout, MBException) as _:
        show_messagebox(pos_id, "$ERROR_COUPON_INTERNAL_ERROR")
        logger.exception("Error sending message to nCoupon")
        raise
    except DiscountMinLevelException as _:
        raise
    except ErrorApplyingDiscount as ex:
        message = "$" + ex.error_code
        return False
    except UnexpectedResponse as ex:
        logger.error("Unexpected response from nCoupon", exc_info=ex)
        message = "$ERROR_COUPON_INTERNAL_ERROR"
        return False
    except Exception as ex:
        logger.error("Unknown error applying coupon", exc_info=ex)
        raise
    finally:
        if message:
            show_messagebox(pos_id, message)


def _apply_coupon_and_set_seat(coupon_code, pos_id, seat, table_id):
    # type: (str, str, str, str) -> None

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order_picture = eTree.XML(pos_ot.orderPicture(pos_id))
    order_id = order_picture.get("orderId")
    sale_lines = order_picture.findall(".//SaleLine")
    last_line_number = max([int(x.get("lineNumber")) for x in sale_lines]) if sale_lines else 0
    apply_coupon(pos_id, table_id=table_id, coupon_code=coupon_code)
    _set_custom_properties_for_added_sale_lines(pos_id, pos_ot, order_id, last_line_number, seat)


def _set_custom_properties_for_added_sale_lines(pos_id, pos_ot, order_id, last_line_number, seat):
    order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))
    sale_lines = order_picture.findall(".//SaleLine")
    last_line_number_after_apply_coupon = max([int(x.get("lineNumber")) for x in sale_lines]) if sale_lines else 0
    if last_line_number != last_line_number_after_apply_coupon:
        sale_type = None if order_picture.get("saleTypeDescr") == "EAT_IN" else order_picture.get("saleTypeDescr")

        for sale_line in sale_lines:
            if int(sale_line.get("level")) > 0 or int(sale_line.get("lineNumber")) <= int(last_line_number):
                continue

            set_custom_properties(pos_ot, sale_line, seat, sale_type)


def _build_custom_error(
    data,  # type: Dict
):
    # type: (...) -> CustomError

    message = data["message"]
    return CustomError(
        message=TranslatedMessage(
            pt=message.get("pt"),
            en=message.get("en"),
            es=message.get("es"),
        )
    )


def _get_translated_message(
    pos_id,        # type: str
    custom_error,  # type: CustomError
):
    language = _get_pos_language(pos_id=pos_id)
    if hasattr(custom_error.message, language):
        return getattr(custom_error.message, language)
    only_language = language[0:2]
    if hasattr(custom_error.message, only_language) and getattr(custom_error.message, only_language) is not None:
        return getattr(custom_error.message, only_language)

    return translate_message(
        model=get_model(pos_id),
        message_key="$ERROR_COUPON_CUSTOM_ERROR_NO_LANGUAGE",
    )


def _get_pos_language(
    pos_id,  # type: str
):
    model = get_model(posid=pos_id)
    language = model.find("Language").get("name").lower()
    return language



# numdb.py - module for handling hierarchically organised numbers
#
# Copyright (C) 2010-2023 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Query structured number format files with number properties.

This module contains functions for reading and querying a database that
stores numbers that use a hierarchical format (e.g. ISBN, IBAN, phone
numbers, etc).

To read a database from a file:

>>> with open('tests/numdb-test.dat', 'r') as f:
...     dbfile = read(f)

To split a number:

>>> dbfile.split('01006')
['0', '100', '6']
>>> dbfile.split('902006')
['90', '20', '06']
>>> dbfile.split('909856')
['90', '985', '6']

To split the number and get properties for each part:

>>> import pprint
>>> pprint.pprint(dbfile.info('01006'))
[('0', {'prop1': 'foo'}), ('100', {'prop2': 'bar'}), ('6', {})]
>>> pprint.pprint(dbfile.info('02006'))
[('0', {'prop1': 'foo'}), ('200', {'prop2': 'bar', 'prop3': 'baz'}), ('6', {})]
>>> pprint.pprint(dbfile.info('03456'))
[('0', {'prop1': 'foo'}), ('345', {'prop2': 'bar', 'prop3': 'baz'}), ('6', {})]
>>> pprint.pprint(dbfile.info('902006'))
[('90', {'prop1': 'booz'}), ('20', {'prop2': 'foo'}), ('06', {})]
>>> pprint.pprint(dbfile.info('909856'))
[('90', {'prop1': 'booz'}), ('985', {'prop2': 'fooz'}), ('6', {})]
>>> pprint.pprint(dbfile.info('9889'))
[('98', {'prop1': 'booz'}), ('89', {'prop2': 'foo'})]
>>> pprint.pprint(dbfile.info('633322'))
[('6', {'prop1': 'boo'}), ('333', {'prop2': 'bar', 'prop3': 'baz', 'prop4': 'bla'}), ('22', {})]
>>> pprint.pprint(dbfile.info('1200333'))
[('1', {'prop1': 'foo'}), ('200', {'prop2': 'bar', 'prop3': 'baz'}), ('333', {'prop4': 'bax'})]

"""

import re


_line_re = re.compile(
    r'^(?P<indent> *)'
    r'(?P<ranges>([^-,\s]+(-[^-,\s]+)?)(,[^-,\s]+(-[^-,\s]+)?)*)\s*'
    r'(?P<props>.*)$')
_prop_re = re.compile(
    r'(?P<prop>[0-9a-zA-Z-_]+)="(?P<value>[^"]*)"')

# this is a cache of open databases
_open_databases = {}

# the prefixes attribute of NumDB is structured as follows:
# prefixes = [
#   [ length, low, high, props, children ]
#   ...
# ]
# where children is a prefixes structure in its own right
# (there is no expected ordering within the list)


class NumDB():
    """Number database."""

    def __init__(self):
        """Construct an empty database."""
        self.prefixes = []

    @staticmethod
    def _find(number, prefixes):
        """Lookup the specified number in the list of prefixes, this will
        return basically what info() should return but works recursively."""
        if not number:
            return []
        part = number
        properties = {}
        next_prefixes = []
        # go over prefixes and find matches
        for length, low, high, props, children in prefixes:
            if len(part) >= length and low <= part[:length] <= high:
                # only use information from the shortest match
                if length < len(part):
                    part = part[:length]
                    properties = {}
                    next_prefixes = []
                properties.update(props)
                next_prefixes.extend(children)
        # return first part and recursively find next matches
        return [(part, properties)] + NumDB._find(number[len(part):], next_prefixes)

    def info(self, number):
        """Split the provided number in components and associate properties
        with each component. This returns a tuple of tuples. Each tuple
        consists of a string (a part of the number) and a dict of properties.
        """
        return NumDB._find(number, self.prefixes)

    def split(self, number):
        """Split the provided number in components. This returns a tuple with
        the number of components identified."""
        return [part for part, props in self.info(number)]


def _parse(fp):
    """Read lines of text from the file pointer and generate indent, length,
    low, high, properties tuples."""
    for line in fp:
        # ignore comments
        if line[0] == '#' or line.strip() == '':
            continue  # pragma: no cover (optimisation takes it out)
        # any other line should parse
        match = _line_re.search(line)
        indent = len(match.group('indent'))
        ranges = match.group('ranges')
        props = dict(_prop_re.findall(match.group('props')))
        children = []
        for rnge in ranges.split(','):
            if '-' in rnge:
                low, high = rnge.split('-')
            else:
                low, high = rnge, rnge
            yield indent, len(low), low, high, props, children


def read(fp):
    """Return a new database with the data read from the specified file."""
    last_indent = 0
    db = NumDB()
    stack = {0: db.prefixes}
    for indent, length, low, high, props, children in _parse(fp):
        if indent > last_indent:
            # set our stack location to the last parent entry
            stack[indent] = stack[last_indent][-1][4]
        stack[indent].append([length, low, high, props, children])
        last_indent = indent
    return db


def _get_resource_stream(name):
    """Return a readable file-like object for the resource."""
    try:  # pragma: no cover (Python 3.9 and newer)
        import importlib.resources
        return importlib.resources.files(__package__).joinpath(name).open('rb')
    except (ImportError, AttributeError):  # pragma: no cover (older Python versions)
        import pkg_resources
        return pkg_resources.resource_stream(__name__, name)


def get(name):
    """Open a database with the specified name to perform queries on."""
    if name not in _open_databases:
        import codecs
        reader = codecs.getreader('utf-8')
        with reader(_get_resource_stream(name + '.dat')) as fp:
            _open_databases[name] = read(fp)
    return _open_databases[name]



from sysactions import action, show_messagebox
from application.interactor import TabsInteractor


@action
def show_open_tabs(
        pos_id  # type: int
):
    # type: (...) -> bool

    tabs_interactor = TabsInteractor(int(pos_id))
    opened_tabs = tabs_interactor.get_opened_tabs()
    if not opened_tabs:
        show_messagebox(pos_id, "$EMPTY_OPENED_TABS")
        return True

    tabs_interactor.show_opened_tabs(opened_tabs)

    return True



import logging
from production.command import ChangePathCommand
from production.model import ProdStates, ProductionOrder
from typing import Dict, List

from ._ProductionBox import ProductionBox


class NoAvailablePaths(Exception):
    pass


class OrderRoundRobinBox(ProductionBox):
    def __init__(self, name, sons, order_changer, paths, logger=None):
        super(OrderRoundRobinBox, self).__init__(name, sons, logger)
        self.order_changer = order_changer
        self.paths = paths
        self.paths_dict = {}  # type: Dict[str, OrderRoundRobinBox.Path]
        self.lowest_height_priority = []  # type: List[str]

        index = 0
        for path in paths:
            self.paths_dict[path] = OrderRoundRobinBox.Path(path, index, sons[index])
            self.lowest_height_priority.append(path)
            index += 1

        self.order_path = {}  # type: Dict[int, OrderRoundRobinBox.Path]
        self.stored_orders = []

    def order_modified(self, order):
        self.logger.debug("Box name: {} - Routing order:\n{}\n".format(self.name, order))
        if order.round_robin_path or order.order_id in self.order_path:
            self.logger.debug("Order previously routed")
            if order.order_id in self.order_path:
                current_path_name = self.order_path[order.order_id].name
            else:
                current_path_name = order.round_robin_path

            current_path = self.paths_dict[current_path_name]
            if order.order_id in self.order_path:
                if order.prod_state in (ProdStates.SERVED, ProdStates.INVALID):
                    self.logger.debug("Know SERVED order, removing from internal dictionaries")
                    del self.order_path[order.order_id]
                    current_path.height -= self._get_order_points(current_path.orders[order.order_id])
                    del current_path.orders[order.order_id]
                else:
                    self.logger.debug("Know not SERVED order, updating internal dictionaries")
                    current_path.height -= self._get_order_points(current_path.orders[order.order_id])
                    current_path.height += self._get_order_points(order)
                    current_path.orders[order.order_id] = order
            elif order.prod_state != ProdStates.SERVED:
                self.logger.debug("Unknown not SERVED order, adding to internal dictionaries")
                current_path.height += self._get_order_points(order)
                self.order_path[order.order_id] = current_path
                current_path.orders[order.order_id] = order

            self.publish_order(current_path_name, order)
        else:
            self.logger.debug("Order never routed")
            if self.refreshing:
                self.logger.debug("Refreshing views. Storing order to process later")
                self.stored_orders.append(order)
                return

            if order.prod_state not in (ProdStates.SERVED, ProdStates.INVALID):
                self.logger.debug("Not served order, adding to internal dictionaries")
                try:
                    lowest_height_path = self.get_lowest_height_path()
                    lowest_height_path.orders[order.order_id] = order
                    lowest_height_path.height += self._get_order_points(order)

                    self.add_path_to_root_order(order, lowest_height_path.name)
                    self.order_path[order.order_id] = lowest_height_path
                    self.publish_order(lowest_height_path.name, order)

                except NoAvailablePaths:
                    self.logger.debug("No available paths")
                    order.prod_state = ProdStates.INVALID
                    self.publish_order_to_all_sons(order)
                    return

            else:
                self.logger.debug("Served order, just propagating to sons")
                self.publish_order_to_all_sons(order)

        if self.logger.level >= logging.DEBUG:
            paths = "Paths and priorities:\n"
            for path in self.paths:
                paths += str(self.paths_dict[path]) + "\n"
            for value in self.lowest_height_priority:
                paths += str(value) + "\n"
            self.logger.debug(paths)

    @staticmethod
    def _get_order_points(order):
        return order.points or 0

    @staticmethod
    def remove_round_robin_path(root_order):
        del root_order.round_robin_path
        return True

    def handle_production_command(self, command):
        if not isinstance(command, ChangePathCommand):
            return super(OrderRoundRobinBox, self).handle_production_command(command)

        order_handled = False
        if command.path in self.paths_dict and self.paths_dict[command.path].enabled != command.enabled:
            self.logger.debug("Changing enable state of path {} to {}".format(command.path, command.enabled))
            order_handled = True
            self.paths_dict[command.path].enabled = command.enabled
            if not command.enabled:
                for order in self.paths_dict[command.path].get_order_list():
                    self.order_changer.change_order(order.order_id, self.remove_round_robin_path)

        super(OrderRoundRobinBox, self).handle_production_command(command)
        return order_handled

    def start_refresh(self):
        self.logger.debug("{} - Start Refresh".format(self.name))

        super(OrderRoundRobinBox, self).start_refresh()
        self.order_path.clear()
        for path in self.paths_dict.values():
            path.height = 0
            path.clear_orders()

    def end_refresh(self):
        self.logger.debug("{} - End Refresh".format(self.name))

        super(OrderRoundRobinBox, self).end_refresh()
        for order in self.stored_orders:
            self.order_modified(order)
        self.stored_orders = []

    def __setattr__(self, name, value):
        if name == "sons" and isinstance(value, list):
            if hasattr(self, "paths_dict"):
                for path in self.paths_dict.values():
                    path.son = value[path.index]
            super(OrderRoundRobinBox, self).__setattr__(name, value)
        else:
            super(OrderRoundRobinBox, self).__setattr__(name, value)

    def get_lowest_height_path(self):
        lowest_height_paths = self.get_lowest_height_paths()
        highest_priority_index = min(map(lambda x: self.lowest_height_priority.index(x.name), lowest_height_paths))
        highest_priority_path = self.lowest_height_priority.pop(highest_priority_index)
        self.lowest_height_priority.append(highest_priority_path)
        return self.paths_dict[highest_priority_path]

    def get_lowest_height_paths(self):
        lowest = None
        for path in self.paths_dict.values():
            if path.enabled and (lowest is None or path.height < lowest):
                lowest = path.height

        ret = []
        for path in self.paths_dict.values():
            if path.enabled and path.height == lowest:
                ret.append(path)

        if not ret:
            raise NoAvailablePaths()
        return ret

    def publish_order(self, lowest_height_path, order):
        # type: (str, ProductionOrder) -> None

        invalid_order = order.clone()
        invalid_order.prod_state = ProdStates.INVALID
        if hasattr(invalid_order, "stamped"):
            del invalid_order.stamped

        for path in self.paths_dict.values():
            if path.name == lowest_height_path:
                path.son.order_modified(order)
            else:
                path.son.order_modified(invalid_order)

    def publish_order_to_all_sons(self, order):
        for son in self.sons:
            son.order_modified(order)

    def add_path_to_root_order(self, order, path):
        def change_order(root_order):
            root_order.round_robin_path = path
            return True

        self.order_changer.change_order(order.order_id, change_order)

    class Path(object):
        def __init__(self, name, index, son):
            self.name = name
            self.index = index
            self.son = son
            self.enabled = True
            self.height = 0
            self.orders = {}

        def get_order_list(self):
            return list(self.orders.values())

        def clear_orders(self):
            self.orders = {}

        def __str__(self):
            message = "{}: {{Height: {}, Enabled: {}, Order Counts: {}}}"
            return message.format(self.name, self.height, self.enabled, len(self.orders))

        def __repr__(self):
            return self.__str__()



class NfceCancelerResponse(object):
    Canceled = "135"
    AlreadyCanceled = "573"

    def __init__(self, c_stat, x_motivo, n_prot):
        self.c_stat = c_stat
        self.x_motivo = x_motivo
        self.n_prot = n_prot



from ..useragents import UserAgent
from ..utils import cached_property


class UserAgentMixin(object):
    """Adds a `user_agent` attribute to the request object which
    contains the parsed user agent of the browser that triggered the
    request as a :class:`~werkzeug.useragents.UserAgent` object.
    """

    @cached_property
    def user_agent(self):
        """The current user agent."""
        return UserAgent(self.environ)



from datetime import datetime, time


class ValidationUtil(object):
    @staticmethod
    def is_inside_allowed_window(
            start_time,     # type: time
            end_time        # type: time
    ):
        # type: (...) -> bool

        if start_time == end_time:
            return False

        now = datetime.now().time()
        is_inside_update_window = start_time <= now <= end_time
        return is_inside_update_window



from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class AllServedItemsBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(AllServedItemsBox, self).__init__(name, sons, logger)

    def change_order(self, order):

        for item in order.items:
            if not item.all_has_at_least_one_tag(["served"]):
                return None

        return order



from application.repository.api import APIRepository
from .dto import UnlockVoucherRequest


class UnlockVoucherInteractor(object):
    def __init__(
        self,               # type: UnlockVoucherInteractor
        api_repository,     # type: APIRepository
    ):
        # type: (...) -> None

        self.api_repository = api_repository

    def unlock_benefit(
        self,           # type: UnlockVoucherInteractor
        request,        # type: UnlockVoucherRequest
    ):
        # type: (...) -> None

        self.api_repository.unlock_benefit(
            benefit_id=request.benefit_id,
            order_picture=request.order_picture,
            table_picture=request.table_picture,
        )



"""Doctest module for HTML comparison.

Usage::

   >>> import lxml.html.usedoctest
   >>> # now do your HTML doctests ...

See `lxml.doctestcompare`.
"""

from lxml import doctestcompare

doctestcompare.temp_install(html=True, del_module=__name__)



# ean.py - functions for handling EANs
#
# Copyright (C) 2011-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""EAN (International Article Number).

Module for handling EAN (International Article Number) codes. This
module handles numbers EAN-13, EAN-8, UPC (12-digit) and GTIN (EAN-14) format.

>>> validate('73513537')
'73513537'
>>> validate('978-0-471-11709-4') # EAN-13 format
'9780471117094'
>>> validate('98412345678908') # GTIN format
'98412345678908'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the EAN to the minimal representation. This strips the number
    of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def calc_check_digit(number):
    """Calculate the EAN check digit for 13-digit numbers. The number passed
    should not have the check bit included."""
    return str((10 - sum((3, 1)[i % 2] * int(n)
                         for i, n in enumerate(reversed(number)))) % 10)


def validate(number):
    """Check if the number provided is a valid EAN-13. This checks the length
    and the check bit but does not check whether a known GS1 Prefix and
    company identifier are referenced."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) not in (14, 13, 12, 8):
        raise InvalidLength()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid EAN-13. This checks the length
    and the check bit but does not check whether a known GS1 Prefix and
    company identifier are referenced."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-

import inspect
import time

from datetime import datetime
from logging import Logger
from typing import (
    List,
    Dict,
    Optional,
)

from mw_helper import BaseRepository
from msgbus import MBEasyContext
from persistence import Connection
from model import RuptureItem


class RuptureRepository(BaseRepository):
    def __init__(self, mb_context, logger):
        # type: (MBEasyContext, Logger) -> None
        super(RuptureRepository, self).__init__(mb_context)

        self.logger = logger

    def __getattribute__(self, name):
        returned = object.__getattribute__(self, name)
        if inspect.isfunction(returned) or inspect.ismethod(returned) and returned.__name__ != "execute_with_connection":
            self.logger.info("[RuptureRepository] Call -> {}".format(returned.__name__))
        return returned

    def get_disabled_items(self):
        def inner_func(conn):
            cursor = conn.select("""
                SELECT DISTINCT
                    ri.ProductCode,
                    p.ProductName,
                    pp.ProductCode AS ProductPartCode
                FROM
                    RupturaItens ri
                    LEFT JOIN Product p ON p.ProductCode = ri.productCode
                    LEFT JOIN ProductPart pp ON pp.PartCode = p.ProductCode
                WHERE
                    EnableDate IS NULL
            """)

            items = []
            for x in cursor:
                product_code = x.get_entry("ProductCode")
                product_name = x.get_entry("ProductName")
                product_part_code = x.get_entry("ProductPartCode")
                rupture_item = RuptureItem(product_code, product_name, product_part_code)
                items.append(rupture_item)

            return items

        return self.execute_with_connection(inner_func)

    def get_enabled_items(self):
        def inner_func(conn):
            cursor = conn.select("""
                SELECT DISTINCT
                    p.ProductCode,
                    p.ProductName,
                    pp.ProductCode AS ProductPartCode
                FROM 
                    ProductKernelParams pkp
                    INNER JOIN Product p ON p.ProductCode = pkp.ProductCode
                    LEFT JOIN RupturaItens ri ON ri.ProductCode = pkp.ProductCode
                    LEFT JOIN ProductPart pp ON pp.PartCode = p.ProductCode
                WHERE pkp.ProductType in (0, 2) AND
                    p.ProductCode NOT IN (SELECT ProductCode FROM RupturaItens WHERE EnableDate IS NULL)
            """)

            items = []
            for x in cursor:
                product_code = x.get_entry("ProductCode")
                product_name = x.get_entry("ProductName")
                product_part_code = x.get_entry("ProductPartCode")
                rupture_item = RuptureItem(product_code, product_name, product_part_code)
                items.append(rupture_item)

            return items

        return self.execute_with_connection(inner_func)

    def get_available_daily_dishes_items(
        self,
        exclude_daily_dishes    # type: List[str]
    ):
        # type: (...) -> List[RuptureItem]

        exclude_query = ", ".join(exclude_daily_dishes)
        if exclude_daily_dishes:
            exclude_query = "AND p.ProductCode NOT IN ({})".format(exclude_query)

        def inner_func(conn):
            cursor = conn.select("""
                SELECT DISTINCT
                    p.ProductCode,
                    p.ProductName
                FROM 
                    Product p
                    INNER JOIN ProductKernelParams pkp ON p.ProductCode = pkp.ProductCode
                WHERE
                    pkp.ProductType in (0, 2) 
                    AND p.ProductCode not in (
                        SELECT ProductCode 
                        FROM ProductCustomParams
                        WHERE CustomParamId == 'INGREDIENT' 
                        AND CustomParamValue == 'true' COLLATE NOCASE
                    )
                    {}
                """.format(exclude_query))

            items = []
            for x in cursor:
                product_code = x.get_entry("ProductCode")
                product_name = x.get_entry("ProductName")
                rupture_item = RuptureItem(
                    product_code=product_code,
                    product_name=product_name,
                    product_part_code=""
                )
                items.append(rupture_item)

            return items

        return self.execute_with_connection(inner_func)

    def get_daily_dishes_items(
        self
    ):
        # type: () -> List[RuptureItem]

        def inner_func(conn):
            cursor = conn.select("""
                SELECT
                    r.ProductCode,
                    r.Quantity,
                    p.ProductName
                FROM 
                    RuptureByQuantity r
                    INNER JOIN Product p ON p.ProductCode = r.ProductCode
                WHERE
                    r.Quantity is not NULL AND
                    r.RemoveDate is NULL AND
                    r.RemoveUser is NULL
                """)

            items = []
            for x in cursor:
                product_code = x.get_entry("ProductCode")
                product_name = x.get_entry("ProductName")
                product_quantity = x.get_entry("Quantity")
                rupture_item = RuptureItem(
                    product_code=product_code,
                    product_name=product_name,
                    product_part_code="",
                    product_quantity=product_quantity
                )
                items.append(rupture_item)

            return items

        return self.execute_with_connection(inner_func)

    def insert_rupture_update(self, session_id, items_to_disable, items_to_enable, disabled_items):
        def inner_func(conn):
            query = ""
            query += self.fill_items_to_disable(session_id, items_to_disable, disabled_items)
            query += self.fill_items_to_enable(session_id, items_to_enable)

            if not query:
                return

            conn.query(query)

        self.execute_with_connection(inner_func)

    def insert_rupture_snapshot(self, disabled_items):
        def inner_func(conn):
            inactive_list = ",".join(disabled_items)
            snapshot_id = self._get_new_snapshot_id()

            query = """INSERT INTO RupturaSnapshot (ID, InactiveProdList) VALUES ('{}','{}');"""
            conn.query(query.format(snapshot_id, inactive_list))
            return snapshot_id

        return self.execute_with_connection(inner_func)

    def insert_rupture_current_state(self, snapshot_id):
        def inner_func(conn):
            query = """INSERT INTO RupturaCurrentState VALUES ('{}', 'Delivery', '999');"""
            conn.query(query.format(snapshot_id))

        self.execute_with_connection(inner_func)

    def insert_event_state(self, event_data):
        def inner_func(conn):
            query = """INSERT OR REPLACE INTO CleanRuptureEvent VALUES ('{}', 0);"""
            conn.query(query.format(event_data))

        self.execute_with_connection(inner_func)

    def insert_daily_dishes_update(
        self,
        dishes_to_set_quantity,     # type: List[Dict]
        dishes_to_unset_quantity,   # type: List[Dict]
        user_id                     # type: str
    ):
        # type: (...) -> None

        def inner_func(conn):
            query = self.set_dishes_quantity_query(
                dishes_to_update=dishes_to_set_quantity,
                user_id=user_id
            )
            query += self.unset_dishes_quantity_query(
                dishes_to_update=dishes_to_unset_quantity,
                user_id=user_id
            )

            if not query:
                return

            conn.query(query)

        self.execute_with_connection(inner_func)

    def check_if_need_to_clean_ruptures(self):
        def inner_func(conn):
            query = """SELECT * FROM CleanRuptureEvent WHERE Processed = 0;"""
            return [row.get_entry(0) for row in conn.select(query)]

        return self.execute_with_connection(inner_func)

    def check_if_has_open_day_event(
        self,
    ):
        # type: (...) -> List
        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> List
            query = """SELECT * FROM RupturaItens 
            WHERE EnableSessionId IN ('AUTO_CLEAN_EVENT', 'AUTO_CLEAN_OPEN_DAY')
            AND DATE(Period) = DATE('now');"""
            return [row.get_entry(0) for row in conn.select(sql=query)]

        return self.execute_with_connection(method=inner_func)

    def mark_clean_rupture_event_as_processed(self):
        def inner_func(conn):
            conn.query("""UPDATE CleanRuptureEvent SET Processed = 1;""")

        self.execute_with_connection(inner_func)

    def mark_pending_ruptures_as_processed(self):
        def inner_func(conn):
            conn.query("""UPDATE RupturaCurrentState
                          SET Processed = 1
                          WHERE Environment = 'Delivery'
                            AND Processed <> 1;""")

        self.execute_with_connection(inner_func)

    def clean_rupture_items(self, session_id):
        # type: (str) ->()

        def inner_func(conn):
            conn.query("""UPDATE RupturaItens 
                          SET EnableDate=DATETIME('NOW'), EnableSessionId='{0}'
                          WHERE EnableDate IS NULL;""".format(session_id))

        self.execute_with_connection(inner_func)

    def clean_rupture_items_by_btn(self, enable_session_id):
        def inner_func(conn):
            conn.query("""UPDATE RupturaItens 
                          SET EnableDate=DATETIME('NOW'), EnableSessionId='{}' 
                          WHERE EnableDate IS NULL;"""
                       .format(enable_session_id))

        self.execute_with_connection(inner_func)

    @staticmethod
    def fill_items_to_disable(session_id, items_to_disable, disabled_items):
        query = ""
        if items_to_disable:
            disabled_list = list(disabled_items.iterkeys())
            filtered_items = set(map(lambda y: y['product_code'], items_to_disable)).difference(disabled_list)
            filtered_items_to_disable = list(z for z in items_to_disable if z['product_code'] in filtered_items)

            if len(filtered_items_to_disable) > 0:
                query += "INSERT INTO RupturaItens (ProductCode, Period, SessionId) VALUES "
                for item in filtered_items_to_disable:
                    query += "('{0}', DATETIME('NOW'), '{1}'),".format(item.get('product_code'), session_id)
                query = query[:-1] + ";"

        return query

    @staticmethod
    def fill_items_to_enable(session_id, items_to_enable):
        query = ""
        if items_to_enable:
            query += """
                        UPDATE RupturaItens
                        SET EnableDate=DATETIME('NOW'), EnableSessionId='{0}'
                        WHERE ProductCode IN ({1}) AND
                            EnableDate IS NULL AND
                            EnableSessionId IS NULL;
                     """.format(session_id, ",".join([x['product_code'] for x in items_to_enable]))

        return query

    @staticmethod
    def _get_new_snapshot_id():
        return int(time.mktime(datetime.now().timetuple()))

    @staticmethod
    def set_dishes_quantity_query(
        dishes_to_update,   # type: List[Dict]
        user_id             # type: str
    ):
        # type: (...) -> str

        if not dishes_to_update:
            return ""

        query = "INSERT OR REPLACE INTO RuptureByQuantity (ProductCode, Quantity, InsertDate, InsertUser) VALUES "

        for item in dishes_to_update:
            query += "({0}, {1}, DATETIME('NOW'), '{2}'),".format(
                item.get('product_code'),
                item.get('product_quantity'),
                user_id
            )
        query = query[:-1] + ";"

        return query

    @staticmethod
    def unset_dishes_quantity_query(
        dishes_to_update,   # type: List[Dict]
        user_id             # type: str
    ):
        # type: (...) -> str

        if not dishes_to_update:
            return ""

        query = """
                    UPDATE RuptureByQuantity 
                    SET RemoveDate = DATETIME('NOW'), RemoveUser = '{}'
                    WHERE ProductCode IN (
                """.format(user_id)

        for item in dishes_to_update:
            query += "{}, ".format(item.get('product_code'))

        query = query[:-2] + ");"

        return query

    def change_dishes_quantity(
        self,
        product_code,   # type: str
        quantity,       # type: str
        user=None       # type: Optional[str]
    ):
        # type: (...) -> None

        def inner_func(conn):
            query_user = """, RemoveDate = DATETIME('NOW'), RemoveUser = '{}' """.format(user) if user else ""
            query = """
                        UPDATE RuptureByQuantity
                        SET Quantity = {0}{1} 
                        WHERE ProductCode = {2}; 
                    """.format(quantity, query_user, product_code)

            conn.query(query)

        self.execute_with_connection(inner_func)

    def get_all_daily_dishes(
        self
    ):
        # type: (...) -> Dict[str, int]

        def inner_func(conn):
            query = """
                SELECT 
                    ProductCode, 
                    Quantity 
                FROM RuptureByQuantity
                WHERE RemoveDate IS NULL;
            """

            return dict([
                (str(row.get_entry("ProductCode") or ''), int(row.get_entry("Quantity")))
                for row in conn.select(query)
            ])

        return self.execute_with_connection(inner_func)



from typing import Union


class AposA8PrintLine(object):
    def __init__(
        self,
        align,                  # type: int
        text,                   # type: Union[str, unicode]
        command_type,           # type: int
        font_size="small",      # type: str
        bold=False,             # type: bool
    ):
        # type: (...) -> None

        self.align = align
        self.text = text
        self.command_type = command_type
        self.font_size = font_size.lower()
        self.bold = bold



# -*- coding: utf-8 -*-

import base64
import hashlib

from typing import List, Dict, Any
from pos_model import Order
from datetime import datetime
from dateutil import tz

from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys


class InfNfeBuilder(NfceXmlPartBuilder):
    cnf_random = ""
    cdv = ""

    def __init__(self, part_builders, c_uf, cnpf, modelo, serie, versao_ws):
        # type: (List[NfceXmlPartBuilder], unicode, unicode, unicode, int, float) -> None
        self.part_builders = part_builders
        self.c_uf = c_uf
        self.cnpf = cnpf
        self.modelo = modelo
        self.serie = serie
        self.versao_ws = versao_ws

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        inner_xml = ""
        for part_builder in self.part_builders:
            inner_xml += part_builder.build_xml(order, context)

        infnfe_xml = "<infNFe versao=\"%.2f\" Id=\"NFe${CHAVE_NFCE}\">" % (3.1 if self.versao_ws in (1, 3) else 4)
        infnfe_xml += inner_xml
        infnfe_xml += "</infNFe>"

        fiscal_number = context[ContextKeys.fiscal_number]
        emission_type = context[ContextKeys.emission_type]

        nfce_key = self._calculate_chave(order,
                                         infnfe_xml,
                                         fiscal_number,
                                         emission_type,
                                         context)

        context[ContextKeys.nfce_key] = nfce_key

        hash_csrt = self._calculate_hash_csrt(context)

        infnfe_xml = infnfe_xml\
            .replace(u"${CHAVE_NFCE}", nfce_key)\
            .replace(u"${CNF_RANDOM}", context[ContextKeys.random_number])\
            .replace(u"${HASH_CSRT}", hash_csrt)\
            .replace(u"${CDV}", context[ContextKeys.dv])

        return infnfe_xml

    def _calculate_chave(self, order, infnfe_xml, numero_nota, tp_emis, context):
        # type: (Order, unicode, int, int, Dict[unicode, Any]) -> unicode
        data_emissao = self._formata_data(order.states[len(order.states) - 1].timestamp)

        random_number = self._calculate_random_number(infnfe_xml)
        context[ContextKeys.random_number] = random_number

        chave = "{}{}{}{}{}{:>03}{:09}{}{}".format(self.c_uf,
                                                   data_emissao[2:4],
                                                   data_emissao[5:7],
                                                   self.cnpf,
                                                   self.modelo,
                                                   self.serie,
                                                   numero_nota,
                                                   tp_emis,
                                                   random_number)

        dv = self._calcula_dv(chave)

        context[ContextKeys.dv] = dv

        return chave + dv

    @staticmethod
    def _formata_data(data):
        # type: (datetime) -> unicode
        local_zone = tz.tzlocal()
        data = data.replace(tzinfo=local_zone)
        data_str = data.strftime("%Y-%m-%dT%H:%M:%S%z")
        data_str = data_str[:22] + ":" + data_str[22:]
        return data_str

    @staticmethod
    def _calcula_dv(nfe_key):
        # type: (unicode) -> unicode
        weight = 2
        current_sum = 0
        for i in range(len(nfe_key) - 1, -1, -1):
            current_sum += int(nfe_key[i]) * weight
            if weight < 9:
                weight += 1
            else:
                weight = 2
        digit = 11 - (current_sum % 11)
        if digit > 9:
            digit = 0
        return str(digit)

    @staticmethod
    def _calculate_random_number(xml):
        # type: (unicode) -> unicode
        cnf_random_int = int(hashlib.sha256(xml).hexdigest(), 16) % (10 ** 8)
        cnf_random_str = str(cnf_random_int).zfill(8)
        return cnf_random_str

    @staticmethod
    def _calculate_hash_csrt(context):
        # type: (Dict[unicode, Any]) -> unicode
        hash_csrt = ""

        if 'csrt' and 'id_csrt' in context:
            concatened_value = context[ContextKeys.csrt] + context[ContextKeys.nfce_key]
            hash_csrt = base64.b64encode(hashlib.sha1(concatened_value).digest())

        return hash_csrt



import json


class ChangeProdStateRequest(object):
    def __init__(
        self,           # type: str
        order_id,       # type: str
        view,           # type: str
        state,          # type: str
        pickup_info,    # type: json
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.view = view
        self.state = state
        self.pickup_info = json.loads(pickup_info) if pickup_info else pickup_info



import json
from logging import Logger

import requests
from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import \
    ErrorRetrievingPendingUpdate, \
    ErrorDownloadingUpdatePackage, \
    ErrorNotifyingAppliedUpdate
from helper import retry, json_serialize
from requests import Response
from typing import Dict
from application.util import ErrorResponseUtil
from application.util import settings

CATALOG_NOT_FOUND = "$CATALOG_NOT_FOUND"


class CatalogUpdaterAPIRepository(object):

    def __init__(self, logger, configs):
        # type: (Logger, Configurations) -> None

        self.logger = logger
        self.configs = configs
        self.update_type = UpdateType.catalog
        self.local_configs = configs.updaters[self.update_type.name]
        self.response_util = ErrorResponseUtil()
        self.error_response = None

    def retrieve_pending_update(self):
        # type: () -> str

        url = self._get_download_url()
        headers = self._get_app_json_headers()

        response = retry(
                logger=self.logger,
                method_name="retrieve_pending_update",
                n=3,
                sleep_seconds=3,
                trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=10, verify=False),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: self._get_error_response(resp))

        if response is None:
            settings.set_update_control("catalog", self.error_response)
            raise ErrorRetrievingPendingUpdate(self.error_response)

        return json.loads(response.content)

    def get_menu_databases(self, menu_link):
        # type: (str) -> Response

        self._init_message_update()

        package = requests.get(menu_link, allow_redirects=True, verify=False)
        self._check_has_file(package)
        self.logger.info("Package successfully downloaded!")

        return package

    def notify_applied_update(self, database_id, update_id):
        # type: (str, int) -> None

        url = self._get_updated_to_url(update_id)
        headers = self._get_app_json_headers()
        data = {"pos_transaction_id": "{}_{}".format(str(database_id), str(update_id))}

        response = retry(
                logger=self.logger,
                method_name="notify_applied_update",
                n=3,
                sleep_seconds=5,
                trying_to_do=lambda: requests.request("POST", url=url, headers=headers, json=data, timeout=120),
                is_success=lambda resp: resp.status_code == 204,
                success=lambda resp: resp,
                failed=lambda resp: self._get_error_response(resp))

        if response is None:
            raise ErrorNotifyingAppliedUpdate(self.error_response)

    def _get_app_json_headers(self):
        # type: () -> Dict[str, str]

        return {
            "X-Api-Key": self.configs.api_key,
            "Content-type": "application/json",
            "Accept": "text/plain"
        }

    def _get_app_zip_headers(self):
        # type: () -> Dict[str, str]

        return {"X-Api-Key": self.configs.api_key}

    def _get_latest_version_url(self):
        # type: () -> str

        end_point_url = self.local_configs.endpoints.get_latest_version.format(storeId=self.configs.store_id)
        return self.configs.base_url + end_point_url

    def _get_download_url(self):
        # type: () -> str

        end_point_url = self.local_configs.endpoints.download.format(storeId=self.configs.store_id)
        return self.configs.base_url + end_point_url

    def _get_updated_to_url(self, updated_id):
        # type: (int) -> str

        end_point_url = self.local_configs.endpoints.update_to.format(
            storeId=self.configs.store_id,
            menu_id=str(updated_id)
        )
        return self.configs.base_url + end_point_url

    def _get_error_response(self, resp):
        # type: (Response) -> None

        self.logger.error("Error downloading catalog: {}".format(json_serialize(resp)))
        self.error_response = self.response_util.get_error_response(resp, CATALOG_NOT_FOUND)

    @staticmethod
    def _check_has_file(package):
        # type: (Response) -> None

        if package is not None and package.status_code == 403:
            settings.set_update_control("catalog", CATALOG_NOT_FOUND)
            raise ErrorDownloadingUpdatePackage(CATALOG_NOT_FOUND)

    @staticmethod
    def _init_message_update():
        settings.set_update_control("catalog", None)



from emptysearchexception import EmptySearchException
from voidorderexception import VoidOrderException
from ._FiscalException import FiscalIncrementException, FiscalGetIdException, FiscalParametersException
from ._StoreNotOperatorWithSatException import StoreNotOperatorWithSatException



"""
WSGI Protocol Linter
====================

This module provides a middleware that performs sanity checks on the
behavior of the WSGI server and application. It checks that the
:pep:`3333` WSGI spec is properly implemented. It also warns on some
common HTTP errors such as non-empty responses for 304 status codes.

.. autoclass:: LintMiddleware

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""
from warnings import warn

from .._compat import implements_iterator
from .._compat import PY2
from .._compat import string_types
from ..datastructures import Headers
from ..http import is_entity_header
from ..wsgi import FileWrapper

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


class WSGIWarning(Warning):
    """Warning class for WSGI warnings."""


class HTTPWarning(Warning):
    """Warning class for HTTP warnings."""


def check_string(context, obj, stacklevel=3):
    if type(obj) is not str:
        warn(
            "'%s' requires strings, got '%s'" % (context, type(obj).__name__),
            WSGIWarning,
        )


class InputStream(object):
    def __init__(self, stream):
        self._stream = stream

    def read(self, *args):
        if len(args) == 0:
            warn(
                "WSGI does not guarantee an EOF marker on the input stream, thus making"
                " calls to 'wsgi.input.read()' unsafe. Conforming servers may never"
                " return from this call.",
                WSGIWarning,
                stacklevel=2,
            )
        elif len(args) != 1:
            warn(
                "Too many parameters passed to 'wsgi.input.read()'.",
                WSGIWarning,
                stacklevel=2,
            )
        return self._stream.read(*args)

    def readline(self, *args):
        if len(args) == 0:
            warn(
                "Calls to 'wsgi.input.readline()' without arguments are unsafe. Use"
                " 'wsgi.input.read()' instead.",
                WSGIWarning,
                stacklevel=2,
            )
        elif len(args) == 1:
            warn(
                "'wsgi.input.readline()' was called with a size hint. WSGI does not"
                " support this, although it's available on all major servers.",
                WSGIWarning,
                stacklevel=2,
            )
        else:
            raise TypeError("Too many arguments passed to 'wsgi.input.readline()'.")
        return self._stream.readline(*args)

    def __iter__(self):
        try:
            return iter(self._stream)
        except TypeError:
            warn("'wsgi.input' is not iterable.", WSGIWarning, stacklevel=2)
            return iter(())

    def close(self):
        warn("The application closed the input stream!", WSGIWarning, stacklevel=2)
        self._stream.close()


class ErrorStream(object):
    def __init__(self, stream):
        self._stream = stream

    def write(self, s):
        check_string("wsgi.error.write()", s)
        self._stream.write(s)

    def flush(self):
        self._stream.flush()

    def writelines(self, seq):
        for line in seq:
            self.write(line)

    def close(self):
        warn("The application closed the error stream!", WSGIWarning, stacklevel=2)
        self._stream.close()


class GuardedWrite(object):
    def __init__(self, write, chunks):
        self._write = write
        self._chunks = chunks

    def __call__(self, s):
        check_string("write()", s)
        self._write.write(s)
        self._chunks.append(len(s))


@implements_iterator
class GuardedIterator(object):
    def __init__(self, iterator, headers_set, chunks):
        self._iterator = iterator
        if PY2:
            self._next = iter(iterator).next
        else:
            self._next = iter(iterator).__next__
        self.closed = False
        self.headers_set = headers_set
        self.chunks = chunks

    def __iter__(self):
        return self

    def __next__(self):
        if self.closed:
            warn("Iterated over closed 'app_iter'.", WSGIWarning, stacklevel=2)

        rv = self._next()

        if not self.headers_set:
            warn(
                "The application returned before it started the response.",
                WSGIWarning,
                stacklevel=2,
            )

        check_string("application iterator items", rv)
        self.chunks.append(len(rv))
        return rv

    def close(self):
        self.closed = True

        if hasattr(self._iterator, "close"):
            self._iterator.close()

        if self.headers_set:
            status_code, headers = self.headers_set
            bytes_sent = sum(self.chunks)
            content_length = headers.get("content-length", type=int)

            if status_code == 304:
                for key, _value in headers:
                    key = key.lower()
                    if key not in ("expires", "content-location") and is_entity_header(
                        key
                    ):
                        warn(
                            "Entity header %r found in 304 response." % key, HTTPWarning
                        )
                if bytes_sent:
                    warn("304 responses must not have a body.", HTTPWarning)
            elif 100 <= status_code < 200 or status_code == 204:
                if content_length != 0:
                    warn(
                        "%r responses must have an empty content length." % status_code,
                        HTTPWarning,
                    )
                if bytes_sent:
                    warn(
                        "%r responses must not have a body." % status_code, HTTPWarning
                    )
            elif content_length is not None and content_length != bytes_sent:
                warn(
                    "Content-Length and the number of bytes sent to the client do not"
                    " match.",
                    WSGIWarning,
                )

    def __del__(self):
        if not self.closed:
            try:
                warn(
                    "Iterator was garbage collected before it was closed.", WSGIWarning
                )
            except Exception:
                pass


class LintMiddleware(object):
    """Warns about common errors in the WSGI and HTTP behavior of the
    server and wrapped application. Some of the issues it check are:

    -   invalid status codes
    -   non-bytestrings sent to the WSGI server
    -   strings returned from the WSGI application
    -   non-empty conditional responses
    -   unquoted etags
    -   relative URLs in the Location header
    -   unsafe calls to wsgi.input
    -   unclosed iterators

    Error information is emitted using the :mod:`warnings` module.

    :param app: The WSGI application to wrap.

    .. code-block:: python

        from werkzeug.middleware.lint import LintMiddleware
        app = LintMiddleware(app)
    """

    def __init__(self, app):
        self.app = app

    def check_environ(self, environ):
        if type(environ) is not dict:
            warn(
                "WSGI environment is not a standard Python dict.",
                WSGIWarning,
                stacklevel=4,
            )
        for key in (
            "REQUEST_METHOD",
            "SERVER_NAME",
            "SERVER_PORT",
            "wsgi.version",
            "wsgi.input",
            "wsgi.errors",
            "wsgi.multithread",
            "wsgi.multiprocess",
            "wsgi.run_once",
        ):
            if key not in environ:
                warn(
                    "Required environment key %r not found" % key,
                    WSGIWarning,
                    stacklevel=3,
                )
        if environ["wsgi.version"] != (1, 0):
            warn("Environ is not a WSGI 1.0 environ.", WSGIWarning, stacklevel=3)

        script_name = environ.get("SCRIPT_NAME", "")
        path_info = environ.get("PATH_INFO", "")

        if script_name and script_name[0] != "/":
            warn(
                "'SCRIPT_NAME' does not start with a slash: %r" % script_name,
                WSGIWarning,
                stacklevel=3,
            )

        if path_info and path_info[0] != "/":
            warn(
                "'PATH_INFO' does not start with a slash: %r" % path_info,
                WSGIWarning,
                stacklevel=3,
            )

    def check_start_response(self, status, headers, exc_info):
        check_string("status", status)
        status_code = status.split(None, 1)[0]

        if len(status_code) != 3 or not status_code.isdigit():
            warn(WSGIWarning("Status code must be three digits"), stacklevel=3)

        if len(status) < 4 or status[3] != " ":
            warn(
                WSGIWarning(
                    "Invalid value for status %r.  Valid "
                    "status strings are three digits, a space "
                    "and a status explanation"
                ),
                stacklevel=3,
            )

        status_code = int(status_code)

        if status_code < 100:
            warn(WSGIWarning("status code < 100 detected"), stacklevel=3)

        if type(headers) is not list:
            warn(WSGIWarning("header list is not a list"), stacklevel=3)

        for item in headers:
            if type(item) is not tuple or len(item) != 2:
                warn(WSGIWarning("Headers must tuple 2-item tuples"), stacklevel=3)
            name, value = item
            if type(name) is not str or type(value) is not str:
                warn(WSGIWarning("header items must be strings"), stacklevel=3)
            if name.lower() == "status":
                warn(
                    WSGIWarning(
                        "The status header is not supported due to "
                        "conflicts with the CGI spec."
                    ),
                    stacklevel=3,
                )

        if exc_info is not None and not isinstance(exc_info, tuple):
            warn(WSGIWarning("invalid value for exc_info"), stacklevel=3)

        headers = Headers(headers)
        self.check_headers(headers)

        return status_code, headers

    def check_headers(self, headers):
        etag = headers.get("etag")

        if etag is not None:
            if etag.startswith(("W/", "w/")):
                if etag.startswith("w/"):
                    warn(
                        HTTPWarning("weak etag indicator should be upcase."),
                        stacklevel=4,
                    )

                etag = etag[2:]

            if not (etag[:1] == etag[-1:] == '"'):
                warn(HTTPWarning("unquoted etag emitted."), stacklevel=4)

        location = headers.get("location")

        if location is not None:
            if not urlparse(location).netloc:
                warn(
                    HTTPWarning("absolute URLs required for location header"),
                    stacklevel=4,
                )

    def check_iterator(self, app_iter):
        if isinstance(app_iter, string_types):
            warn(
                "The application returned astring. The response will send one character"
                " at a time to the client, which will kill performance. Return a list"
                " or iterable instead.",
                WSGIWarning,
                stacklevel=3,
            )

    def __call__(self, *args, **kwargs):
        if len(args) != 2:
            warn("A WSGI app takes two arguments.", WSGIWarning, stacklevel=2)

        if kwargs:
            warn(
                "A WSGI app does not take keyword arguments.", WSGIWarning, stacklevel=2
            )

        environ, start_response = args

        self.check_environ(environ)
        environ["wsgi.input"] = InputStream(environ["wsgi.input"])
        environ["wsgi.errors"] = ErrorStream(environ["wsgi.errors"])

        # Hook our own file wrapper in so that applications will always
        # iterate to the end and we can check the content length.
        environ["wsgi.file_wrapper"] = FileWrapper

        headers_set = []
        chunks = []

        def checking_start_response(*args, **kwargs):
            if len(args) not in (2, 3):
                warn(
                    "Invalid number of arguments: %s, expected 2 or 3." % len(args),
                    WSGIWarning,
                    stacklevel=2,
                )

            if kwargs:
                warn("'start_response' does not take keyword arguments.", WSGIWarning)

            status, headers = args[:2]

            if len(args) == 3:
                exc_info = args[2]
            else:
                exc_info = None

            headers_set[:] = self.check_start_response(status, headers, exc_info)
            return GuardedWrite(start_response(status, headers, exc_info), chunks)

        app_iter = self.app(environ, checking_start_response)
        self.check_iterator(app_iter)
        return GuardedIterator(app_iter, headers_set, chunks)



from typing import Optional


class Store(object):
    def __init__(
            self,
            crt=None,                   # type: Optional[int]
            name=None,                  # type: Optional[unicode]
            cnpj=None,                  # type: Optional[unicode]
            state_registration=None,    # type: Optional[unicode]
            postal_code=None,           # type: Optional[unicode]
            address_street=None,        # type: Optional[unicode]
            address_number=None,        # type: Optional[unicode]
            address_complement=None,    # type: Optional[unicode]
            neighborhood=None,          # type: Optional[unicode]
            city=None,                  # type: Optional[unicode]
            state=None,                 # type: Optional[unicode]
            city_code=None,             # type: Optional[int]
            state_code=None,            # type: Optional[int]
    ):

        # type: (...) -> None

        self.name = name
        self.crt = crt
        self.cnpj = cnpj
        self.state_registration = state_registration
        self.postal_code = postal_code
        self.address_street = address_street
        self.address_number = address_number
        self.address_complement = address_complement
        self.neighborhood = neighborhood
        self.city = city
        self.state = state
        self.city_code = city_code
        self.state_code = state_code



import json
import base64

from typing import Optional


class FinishPaymentData(object):
    def __init__(
        self,
        pos_id,             # type: str
        order_id,           # type: str
        fiscal_date,        # type: str
        fiscal_time,        # type: str
        status,             # type: str
        tender_details,     # type: str
        tender_id=""        # type: Optional[str]
    ):
        # type: (...) -> None

        self.pos_id = int(pos_id)
        self.order_id = int(order_id)
        self.fiscal_date = fiscal_date
        self.fiscal_time = fiscal_time
        self.status = int(status)
        self.tender_details = json.loads(base64.b64decode(tender_details) or "{}")
        self.tender_id = tender_id

    def __repr__(self):
        return """FinishPaymentData(
            pos_id={}, order_id={}, fiscal_date="{}", fiscal_time="{}", status={}, tender_details="{}", tender_id={}
        )""".format(
            self.pos_id,
            self.order_id,
            self.fiscal_date,
            self.fiscal_time,
            self.status,
            json.dumps(self.tender_details),
            self.tender_id
        )

    def __str__(self):
        return self.__repr__()



from xml.etree import cElementTree as eTree
from sysactions import (
    get_model,
    get_posot,
)
from services.model import DiscountAppliedRequest
from externals.repository import AlertsRepository
from services.model import ConfiguredAlert
from services.util import (
    get_configured_alerts_by_type,
    insert_event_into_repository,
    get_order_total,
)


class DiscountAppliedInteractor(object):
    def __init__(
        self,
        alerts_repository,  # type: AlertsRepository
    ):
        # type: (...) -> None

        self.alerts_repository = alerts_repository
        self.configured_alerts = self.alerts_repository.get_configured_alerts()

    def process_request(
        self,
        request,  # type: DiscountAppliedRequest
    ):
        # type: (...) -> None

        configured_alerts = get_configured_alerts_by_type(
            notify_type="DISCOUNT",
            configured_alerts=self.configured_alerts,
        )
        if not configured_alerts:
            return

        pos_id = request.pos_id
        order_id = request.order_id
        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        order_picture = eTree.XML(pos_ot.orderPicture(orderid=order_id)).find("Order")

        discount_amount = float(order_picture.get("discountAmount"))
        total = get_order_total(order_picture=order_picture)
        discount_percentage = (discount_amount / total) * 100.0

        order_picture_dict = None
        table_picture_dict = None
        for alert in configured_alerts:
            valid_conditions = self._validate_alert(
                alert=alert,
                discount_amount=discount_amount,
                discount_percentage=discount_percentage,
            )
            if not valid_conditions:
                continue

            order_picture_dict, table_picture_dict = insert_event_into_repository(
                pos_id=pos_id,
                alerts_repository=self.alerts_repository,
                order_picture=order_picture,
                alert=alert,
                order_picture_dict=order_picture_dict,
                table_picture_dict=table_picture_dict,
            )

    @staticmethod
    def _validate_alert(
        alert,                # type: ConfiguredAlert
        discount_amount,      # type: float
        discount_percentage,  # type: float
    ):
        # type: (...) -> bool

        alert_conditions = alert.alert_json.get("conditions")

        has_discount_alert = "discountAmount" in alert_conditions
        discount_reached_alert_value = discount_amount >= float(alert_conditions.get("discountAmount"))
        if has_discount_alert and not discount_reached_alert_value:
            return False

        has_discount_alert = "discountPercentage" in alert_conditions
        discount_reached_alert_value = discount_percentage >= float(alert_conditions.get("discountPercentage"))
        if has_discount_alert and not discount_reached_alert_value:
            return False

        return True



# -*- coding: utf-8 -*-
import datetime
import json
import sysactions

from plistlib import Dict
from _AposA8PrintLine import AposA8PrintLine
from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType
from _AposA8FiscalPrinter import AposA8FiscalPrinter
from common import PrinterCommon
from dateutil import tz
from msgbus import MBEasyContext
from _AposA8Helpers import (
    APOS_DASH,
    APOS_NEWLINE,
    append_tax_data,
    split_lines,
    append_lines_from_array,
    build_call_id,
)
from xml.etree import cElementTree as eTree

from old_helper import remove_xml_namespace
from clichehelper import ClicheHelper
from typing import (
    Optional,
    List,
)

from _AposA8FontSize import AposA8FontSize
from models import LoyaltyConfigurations


class NfceAposA8FiscalPrinter(AposA8FiscalPrinter):
    def __init__(
        self,
        mb_context,                    # type: MBEasyContext
        order_ref_ids,                 # type: list
        cliche_helper,                 # type: ClicheHelper
        qrcode_url,                    # type: str
        customer_text_field,           # type: str
        printer_common,                # type: PrinterCommon
        coupon_fiscal_parameters,      # type: Dict[str, bool]
        font_size_withdrawal_panel,    # type: str
        danfe_font_size_mobile,        # type: str
        order_ref_position,            # type: str
        loyalty_configurations,        # type: LoyaltyConfigurations
    ):
        # type: (...) -> None

        super(NfceAposA8FiscalPrinter, self).__init__(
            mb_context=mb_context,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            customer_text_field=customer_text_field,
            coupon_fiscal_parameters=coupon_fiscal_parameters,
            font_size_withdrawal_panel=font_size_withdrawal_panel,
            danfe_font_size_mobile=danfe_font_size_mobile,
            order_ref_position=order_ref_position,
            loyalty_configurations=loyalty_configurations,
        )

        self.qrcode_check_url = qrcode_url
        self.satisfaction_survey_text = None
        self.customer_text_field = customer_text_field
        self.printer_common = printer_common
        self.danfe_font_size_mobile = danfe_font_size_mobile

    def print_fiscal(
        self,               # type: NfceAposA8FiscalPrinter
        pos_id,             # type: str
        order,              # type: eTree.Element
        tenders,            # type: list
        fiscal_request,     # type: eTree.Element
        fiscal_response,    # type: eTree.Element
        total_taxes,        # type: dict
        url_sefaz="",       # type: str
        is_reprint=False,   # type: Optional[bool]
    ):
        # type: (...) -> str

        model = sysactions.get_model(pos_id)
        fiscal_xml = remove_xml_namespace(eTree.tostring(fiscal_request))
        is_contingency = fiscal_xml.find(".//tpEmis").text == "9"

        if is_reprint and is_contingency and fiscal_response is not None:
            is_contingency = False

        report = []

        if is_contingency and not is_reprint:
            report = self._fiscal_report(
                order=order,
                tenders=tenders,
                fiscal_request=fiscal_request,
                fiscal_response=fiscal_response,
                total_taxes=total_taxes,
                via="Via Estabelecimento",
                url_sefaz=self.qrcode_check_url,
                model=model,
                is_contingency=is_contingency,
            )
            report.append(APOS_NEWLINE)

        via = "Via Consumidor" if is_reprint or is_contingency else ""

        report.extend(
            self._fiscal_report(
                order=order,
                tenders=tenders,
                fiscal_request=fiscal_request,
                fiscal_response=fiscal_response,
                total_taxes=total_taxes,
                via=via,
                url_sefaz=self.qrcode_check_url,
                model=model,
                is_contingency=is_contingency,
            )
        )

        print_loyalty_coupon = PrinterCommon.should_print_loyalty_coupon(
            order=order,
            loyalty_configurations=self.loyalty_configurations,
        )
        if print_loyalty_coupon:
            report.extend(
                self._print_loyalty(
                    pos_id=pos_id,
                    order=order,
                    fiscal_request=fiscal_request,
                )
            )

        return json.dumps(report, default=lambda x: x.__dict__)

    def print_cancel(self, pos_id, data):
        # type: (str, str) -> None
        pass

    def _print_loyalty(
        self,                    # type: AposA8FiscalPrinter
        pos_id,                  # type: str
        order,                   # type: eTree.Element
        fiscal_request,          # type: eTree.Element
    ):
        # type: (...) -> List

        protocol_date, qr_code_text = PrinterCommon.get_nfce_params_to_print_loyalty_coupon(
            fiscal_request=fiscal_request,
            order=order,
        )
        report = self._build_loyalty_report(
            model=sysactions.get_model(posid=pos_id),
            qr_code_text=qr_code_text,
            protocol_date=protocol_date,
        )

        return report

    def _fiscal_report(
        self,
        order,                  # type: eTree.Element
        tenders,                # type: list
        fiscal_request,         # type: eTree.Element
        fiscal_response,        # type: eTree.Element
        total_taxes,            # type: dict
        via,                    # type: str
        url_sefaz,              # type: str
        model=None,             # type: Optional[eTree.Element]
        is_contingency=False,   # type: Optional[bool]
    ):
        # type: (...) -> list

        report = []

        if self.order_ref_position == "BEFORE_FISCAL_DATA":
            build_call_id(
                model=model,
                report=report,
                order=order,
                font_size_withdrawal_panel=self.font_size_withdrawal_panel,
                printer_common=self.printer_common,
                order_ref_ids=self.order_ref_ids,
                position="before",
            )

        self._append_cliche(report=report)

        if is_contingency:
            report.append(APOS_NEWLINE)
            self._append_contingency_warning(report=report)
            report.append(APOS_DASH)

        self._append_coupon_body(
            model=model,
            order=order,
            report=report,
            tenders=tenders,
            printer_common=self.printer_common,
        )

        fiscal_xml = remove_xml_namespace(eTree.tostring(fiscal_request))

        self._append_sefaz_url_and_key(
            report=report,
            fiscal_xml=fiscal_xml,
            url_sefaz=url_sefaz,
        )

        self._append_cpf_cnpj(
            report=report,
            fiscal_xml=fiscal_xml,
        )

        self._append_nfce_data(
            report=report,
            fiscal_xml=fiscal_xml,
        )

        if is_contingency:
            report.append(self._get_text_line(text=via))

        response_xml = None
        if fiscal_response is not None:
            response_xml = remove_xml_namespace(eTree.tostring(fiscal_response))

        self._append_auth_protocol(
            report=report,
            response_xml=response_xml,
        )

        self._append_homolog_warning(
            report=report,
            fiscal_xml=fiscal_xml,
        )

        if is_contingency:
            self._append_contingency_warning(report=report)

        self._append_qr_code(
            report=report,
            fiscal_xml=fiscal_xml,
        )

        self._append_response_protocol_information(
            report=report,
            response_xml=response_xml,
        )

        append_tax_data(
            report=report,
            total_taxes=total_taxes,
            font_size=self.danfe_font_size_mobile,
        )

        self._append_customer_text_field(report=report)

        return report

    def _append_response_protocol_information(
        self,
        report,  # type: List[AposA8PrintLine]
        response_xml,  # type: eTree
    ):
        # type: (...) -> None

        if response_xml is not None:
            message_code_tag = response_xml.find(".//infProt/cMsg")
            message_code = message_code_tag.text if message_code_tag is not None else None
            message_text_tag = response_xml.find(".//infProt/xMsg")
            message_text = message_text_tag.text if message_text_tag is not None else None
            if message_code == '200' and message_text not in [None, '', ' ']:
                text_to_print = message_text.replace('|', '\n').strip()
                report.append(self._get_text_line(text=text_to_print))

    def _append_auth_protocol(
        self,
        report,             # type: List[AposA8PrintLine]
        response_xml,       # type: eTree
    ):
        # type: (...) -> None

        if response_xml is not None:
            protocolo_element = response_xml.find(".//nProt").text
            if protocolo_element is not None:
                auth_protocol_text = "Protocolo de autorizacao: " + protocolo_element
                report.append(self._get_text_line(text=auth_protocol_text))

                protocol_date = response_xml.find(".//dhRecbto").text
                protocol_date_text = "Data de autorizacao: " + self._formata_data(protocol_date)
                report.append(self._get_text_line(text=protocol_date_text))

    def _append_homolog_warning(
        self,
        report,         # type: List[AposA8PrintLine]
        fiscal_xml,     # type: eTree
    ):
        # type: (...) -> None

        tp_amb = fiscal_xml.find(".//tpAmb").text
        if tp_amb == "2":
            text_to_print = ["EMITIDA EM HOMOLOGACAO - SEM VALOR FISCAL"]

            if self.danfe_font_size_mobile != AposA8FontSize.SMALL:
                text_to_print = ["EMITIDA EM HOMOLOGACAO", 'SEM VALOR FISCAL']

            for text in text_to_print:
                report.append(self._get_text_line(text=text))

    @staticmethod
    def _append_qr_code(report, fiscal_xml):
        url = fiscal_xml.find(".//qrCode").text.strip()
        report.append(AposA8PrintLine(AposA8Align.CENTER, url, AposA8CommandType.PRINT_QR_CODE))

    def _append_nfce_data(
        self,
        report,         # type: List[AposA8PrintLine]
        fiscal_xml,     # type: eTree
    ):
        # type: (...) -> None

        numero = fiscal_xml.find(".//nNF").text
        serie = fiscal_xml.find(".//serie").text
        data_emissao_xml = fiscal_xml.find(".//dhEmi").text
        data_emissao_str = self._formata_data(data_emissao_xml)
        nfce_numero = "NFC-e Numero %09d " % (int(numero))
        nfce_serie = "Serie %03d " % int(serie)
        nfce_data_emissao = "%s" % data_emissao_str

        font_is_small = True if self.danfe_font_size_mobile == AposA8FontSize.SMALL else False

        small_lines = ["{}{}".format(nfce_numero, nfce_serie), nfce_data_emissao]
        large_lines = [nfce_numero, "{}{}".format(nfce_serie, nfce_data_emissao)]
        lines = small_lines if font_is_small else large_lines

        append_lines_from_array(
            report=report,
            lines=lines,
            font_size=self.danfe_font_size_mobile,
        )

    def _append_cpf_cnpj(
        self,
        report,         # type: List[AposA8PrintLine]
        fiscal_xml,     # type: eTree
    ):
        # type: (...) -> None

        cpf_cnpj = ""
        cpf = fiscal_xml.find(".//dest/CPF")
        if cpf is not None:
            cpf_cnpj = cpf.text
        else:
            cnpj = fiscal_xml.find(".//dest/CNPJ")
            if cnpj is not None:
                cpf_cnpj = cnpj.text
        cpf_cnpj_text = "CONSUMIDOR %s" % ("NAO INFORMADO" if not cpf_cnpj else ("- CPF " + cpf_cnpj))

        report.append(self._get_text_line(text=cpf_cnpj_text))

    def _append_sefaz_url_and_key(
        self,
        report,		    # type: List[AposA8PrintLine]
        fiscal_xml,     # type: eTree
        url_sefaz,      # type: str
    ):
        # type: (...) -> None

        text_to_print = "Consulte pela Chave de Acesso em"
        report.append(self._get_text_line(text=text_to_print))
        split_lines(
            report=report,
            text=url_sefaz,
            font_size=self.danfe_font_size_mobile,
            separator="/",
        )

        self._append_danfe_key(
            report=report,
            key=fiscal_xml.find(".//infNFe").attrib['Id'][3:],
        )

    def _append_contingency_warning(
        self,
        report,		# type: List[AposA8PrintLine]
    ):
        # type: (...) -> None

        report.append(self._get_text_line(text="EMITIDA EM CONTINGENCIA"))
        report.append(self._get_text_line(text="Pendente de autorizacao"))

    def _append_header(
        self,
        report,		# type: List[AposA8PrintLine]
    ):
        # type: (...) -> None

        text = "Documento Auxiliar da Nota Fiscal de Consumidor Eletronica"
        split_lines(
            report=report,
            text=text,
            font_size=self.danfe_font_size_mobile,
        )
        report.append(APOS_DASH)

    @staticmethod
    def _formata_data(data, mascara="%d/%m/%Y %H:%M:%S"):
        data_emissao_date_utc = data[:19]
        data_emissao_date_utc = datetime.datetime.strptime(data_emissao_date_utc, "%Y-%m-%dT%H:%M:%S")
        data_emissao_offset = int(data[20:22]) * 60 + int(data[23:])
        if data[19] == "+":
            data_emissao_offset *= -1
        data_emissao_date_utc = data_emissao_date_utc + datetime.timedelta(minutes=data_emissao_offset)
        from_zone = tz.tzutc()
        to_zone = tz.tzlocal()

        data_emissao_date = data_emissao_date_utc.replace(tzinfo=from_zone)
        # Convert time zone
        data_emissao_date = data_emissao_date.astimezone(to_zone)
        data_emissao_str = data_emissao_date.strftime(mascara)
        return data_emissao_str



from xml.etree import cElementTree as eTree


def get_order_total(
    order_picture,  # type: eTree.XML
):
    # type: (...) -> float

    total = float(order_picture.get("totalGross"))

    if order_picture.get("discountAmount") is not None:
        total += float(order_picture.get("discountAmount"))

    if order_picture.get("tip") is not None:
        total += float(order_picture.get("tip"))

    return total



# -*- coding: utf-8 -*-

import csv
import logging
import os
import re

from old_helper import get_sale_line_priced_items, round_half_away_from_zero
from pos_model import SaleItem
from common import FiscalParameterNotFound

logger = logging.getLogger("FiscalWrapper")


class IbptTaxProcessor(object):
    def __init__(self, mbcontext, ibpt_tax_files_path, uf, fiscal_parameter_controller):
        self.mbcontext = mbcontext
        self.ibpt_tax_files_path = ibpt_tax_files_path
        self.uf = uf
        self.ibpt_tax_dict = self._get_ibpt_tax_dict()
        self.fiscal_parameter_controller = fiscal_parameter_controller
        self.tax_types = ['nacionalfederal', 'estadual']

    def _get_ibpt_tax_dict(self):
        if os.path.isdir(self.ibpt_tax_files_path):
            ibpt_file = self._get_csv_file(self.ibpt_tax_files_path, self.uf)
            self.ibpt_tax_dict = self._convert_csv_to_dict(ibpt_file, self.ibpt_tax_files_path)
        else:
            logger.error("IBPT Tax files directory not found. Path: {}".format(self.ibpt_tax_files_path))
            self.ibpt_tax_dict = None

        return self.ibpt_tax_dict

    def ibpt_tax_calculator(self, order):
        if self.ibpt_tax_dict:
            product_list = self._get_order_products_ncm_list(order)
            total_taxes = self._get_total_taxes(product_list)
        else:
            total_taxes = self._get_null_taxes()

        return total_taxes

    def get_item_tax(self, sale_line):
        if self.ibpt_tax_dict:
            product_list = self._get_ncm_by_products([sale_line])
            if not product_list:
                return "0,00"

            total_taxes = self._get_total_taxes(product_list)
        else:
            total_taxes = self._get_null_taxes()

        tax_sum = sum(total_taxes[tax_type]['value'] for tax_type in total_taxes)
        return "{:.02f}".format(tax_sum)

    def _get_order_products_ncm_list(self, order):
        sale_lines = order.findall("SaleLine")
        sale_lines = get_sale_line_priced_items(sale_lines)

        return self._get_ncm_by_products(sale_lines)

    def _get_ncm_by_products(self, sale_lines):
        products = {}
        for sale_line in sale_lines:
            part_code = sale_line.part_code if type(sale_line) == SaleItem else sale_line.attrib['partCode']
            product_name = sale_line.product_name if type(sale_line) == SaleItem else sale_line.attrib['productName']
            try:
                ncm = int(self.fiscal_parameter_controller.get_parameter(part_code, "NCM"))
            except FiscalParameterNotFound:
                logger.warning("NCM not found: {}-{}".format(part_code, product_name.encode("utf-8")))
                continue
            correct_item_price = float(sale_line.item_price if type(sale_line) == SaleItem else sale_line.attrib['correctItemPrice'])
            if ncm not in products:
                products[ncm] = correct_item_price
            else:
                products[ncm] += correct_item_price
        return products

    def _get_total_taxes(self, products):
        total_dict = {}
        for tax_type in self.tax_types:
            total_dict[tax_type] = {}
            total_value = 0
            total_tax_value = 0
            for ncm in products:
                if ncm in self.ibpt_tax_dict:
                    if self.ibpt_tax_dict[ncm][tax_type] in (None, ""):
                        tax_percent = 0
                    else:
                        tax_percent = float(self.ibpt_tax_dict[ncm][tax_type])
                    tax_value = (products[ncm] * tax_percent) / 100
                    total_value += products[ncm]
                    total_tax_value += tax_value

            if total_tax_value == 0 or total_value == 0:
                total_dict[tax_type]['value'] = 0
                total_dict[tax_type]['percent'] = 0
            else:
                total_dict[tax_type]['value'] = round_half_away_from_zero(total_tax_value, 2)
                total_dict[tax_type]['percent'] = round_half_away_from_zero(((total_tax_value * 100) / total_value), 2)

        return total_dict

    def _get_null_taxes(self):
        total_dict = {}
        for tax_type in self.tax_types:
            total_dict[tax_type] = {}
            total_dict[tax_type]['value'] = 0
            total_dict[tax_type]['percent'] = 0
        return total_dict

    @staticmethod
    def _get_csv_file(ibpt_tax_files_path, store_uf):
        csv_file = None
        for tax_file in os.listdir(ibpt_tax_files_path):
            file_uf = re.search(r"(?<=Tax)([^\d]+).*\.csv$", tax_file)
            if file_uf not in (None, "") and file_uf.group(1) == store_uf:
                csv_file = tax_file
                break
        if not csv_file:
            logger.error("IBPT Tax file not found for UF: {}".format(store_uf))
        return csv_file

    @staticmethod
    def _convert_csv_to_dict(csv_file, ibpt_tax_files_path):
        file_dict = None
        if csv_file:
            with open(os.path.join(ibpt_tax_files_path, csv_file), 'r') as f:
                reader = csv.reader(f, delimiter=';', quotechar='"')
                result = [r for r in reader]
                file_dict = {}
                for row in result[1:]:
                    result_line = {}
                    for i, item in enumerate(result[0]):
                        result_line[item] = row[i]
                        file_dict[int(row[0])] = result_line
        return file_dict



class TranslatedMessage(object):
    def __init__(
        self,  # type: TranslatedMessage
        pt,    # type: str
        en,    # type: str
        es,    # type: str
    ):
        # type: (...) -> None

        self.pt = pt
        self.en = en
        self.es = es



# -*- coding: utf-8 -*-
"""
    werkzeug._internal
    ~~~~~~~~~~~~~~~~~~

    This module provides internally used helpers and constants.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import inspect
import logging
import re
import string
from datetime import date
from datetime import datetime
from itertools import chain
from weakref import WeakKeyDictionary

from ._compat import int_to_byte
from ._compat import integer_types
from ._compat import iter_bytes
from ._compat import range_type
from ._compat import text_type


_logger = None
_signature_cache = WeakKeyDictionary()
_epoch_ord = date(1970, 1, 1).toordinal()
_legal_cookie_chars = (
    string.ascii_letters + string.digits + u"/=!#$%&'*+-.^_`|~:"
).encode("ascii")

_cookie_quoting_map = {b",": b"\\054", b";": b"\\073", b'"': b'\\"', b"\\": b"\\\\"}
for _i in chain(range_type(32), range_type(127, 256)):
    _cookie_quoting_map[int_to_byte(_i)] = ("\\%03o" % _i).encode("latin1")

_octal_re = re.compile(br"\\[0-3][0-7][0-7]")
_quote_re = re.compile(br"[\\].")
_legal_cookie_chars_re = br"[\w\d!#%&\'~_`><@,:/\$\*\+\-\.\^\|\)\(\?\}\{\=]"
_cookie_re = re.compile(
    br"""
    (?P<key>[^=;]+)
    (?:\s*=\s*
        (?P<val>
            "(?:[^\\"]|\\.)*" |
             (?:.*?)
        )
    )?
    \s*;
""",
    flags=re.VERBOSE,
)


class _Missing(object):
    def __repr__(self):
        return "no value"

    def __reduce__(self):
        return "_missing"


_missing = _Missing()


def _get_environ(obj):
    env = getattr(obj, "environ", obj)
    assert isinstance(env, dict), (
        "%r is not a WSGI environment (has to be a dict)" % type(obj).__name__
    )
    return env


def _has_level_handler(logger):
    """Check if there is a handler in the logging chain that will handle
    the given logger's effective level.
    """
    level = logger.getEffectiveLevel()
    current = logger

    while current:
        if any(handler.level <= level for handler in current.handlers):
            return True

        if not current.propagate:
            break

        current = current.parent

    return False


def _log(type, message, *args, **kwargs):
    """Log a message to the 'werkzeug' logger.

    The logger is created the first time it is needed. If there is no
    level set, it is set to :data:`logging.INFO`. If there is no handler
    for the logger's effective level, a :class:`logging.StreamHandler`
    is added.
    """
    global _logger

    if _logger is None:
        _logger = logging.getLogger("werkzeug")

        if _logger.level == logging.NOTSET:
            _logger.setLevel(logging.INFO)

        if not _has_level_handler(_logger):
            _logger.addHandler(logging.StreamHandler())

    getattr(_logger, type)(message.rstrip(), *args, **kwargs)


def _parse_signature(func):
    """Return a signature object for the function."""
    if hasattr(func, "im_func"):
        func = func.im_func

    # if we have a cached validator for this function, return it
    parse = _signature_cache.get(func)
    if parse is not None:
        return parse

    # inspect the function signature and collect all the information
    if hasattr(inspect, "getfullargspec"):
        tup = inspect.getfullargspec(func)
    else:
        tup = inspect.getargspec(func)
    positional, vararg_var, kwarg_var, defaults = tup[:4]
    defaults = defaults or ()
    arg_count = len(positional)
    arguments = []
    for idx, name in enumerate(positional):
        if isinstance(name, list):
            raise TypeError(
                "cannot parse functions that unpack tuples in the function signature"
            )
        try:
            default = defaults[idx - arg_count]
        except IndexError:
            param = (name, False, None)
        else:
            param = (name, True, default)
        arguments.append(param)
    arguments = tuple(arguments)

    def parse(args, kwargs):
        new_args = []
        missing = []
        extra = {}

        # consume as many arguments as positional as possible
        for idx, (name, has_default, default) in enumerate(arguments):
            try:
                new_args.append(args[idx])
            except IndexError:
                try:
                    new_args.append(kwargs.pop(name))
                except KeyError:
                    if has_default:
                        new_args.append(default)
                    else:
                        missing.append(name)
            else:
                if name in kwargs:
                    extra[name] = kwargs.pop(name)

        # handle extra arguments
        extra_positional = args[arg_count:]
        if vararg_var is not None:
            new_args.extend(extra_positional)
            extra_positional = ()
        if kwargs and kwarg_var is None:
            extra.update(kwargs)
            kwargs = {}

        return (
            new_args,
            kwargs,
            missing,
            extra,
            extra_positional,
            arguments,
            vararg_var,
            kwarg_var,
        )

    _signature_cache[func] = parse
    return parse


def _date_to_unix(arg):
    """Converts a timetuple, integer or datetime object into the seconds from
    epoch in utc.
    """
    if isinstance(arg, datetime):
        arg = arg.utctimetuple()
    elif isinstance(arg, integer_types + (float,)):
        return int(arg)
    year, month, day, hour, minute, second = arg[:6]
    days = date(year, month, 1).toordinal() - _epoch_ord + day - 1
    hours = days * 24 + hour
    minutes = hours * 60 + minute
    seconds = minutes * 60 + second
    return seconds


class _DictAccessorProperty(object):
    """Baseclass for `environ_property` and `header_property`."""

    read_only = False

    def __init__(
        self,
        name,
        default=None,
        load_func=None,
        dump_func=None,
        read_only=None,
        doc=None,
    ):
        self.name = name
        self.default = default
        self.load_func = load_func
        self.dump_func = dump_func
        if read_only is not None:
            self.read_only = read_only
        self.__doc__ = doc

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        storage = self.lookup(obj)
        if self.name not in storage:
            return self.default
        rv = storage[self.name]
        if self.load_func is not None:
            try:
                rv = self.load_func(rv)
            except (ValueError, TypeError):
                rv = self.default
        return rv

    def __set__(self, obj, value):
        if self.read_only:
            raise AttributeError("read only property")
        if self.dump_func is not None:
            value = self.dump_func(value)
        self.lookup(obj)[self.name] = value

    def __delete__(self, obj):
        if self.read_only:
            raise AttributeError("read only property")
        self.lookup(obj).pop(self.name, None)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)


def _cookie_quote(b):
    buf = bytearray()
    all_legal = True
    _lookup = _cookie_quoting_map.get
    _push = buf.extend

    for char in iter_bytes(b):
        if char not in _legal_cookie_chars:
            all_legal = False
            char = _lookup(char, char)
        _push(char)

    if all_legal:
        return bytes(buf)
    return bytes(b'"' + buf + b'"')


def _cookie_unquote(b):
    if len(b) < 2:
        return b
    if b[:1] != b'"' or b[-1:] != b'"':
        return b

    b = b[1:-1]

    i = 0
    n = len(b)
    rv = bytearray()
    _push = rv.extend

    while 0 <= i < n:
        o_match = _octal_re.search(b, i)
        q_match = _quote_re.search(b, i)
        if not o_match and not q_match:
            rv.extend(b[i:])
            break
        j = k = -1
        if o_match:
            j = o_match.start(0)
        if q_match:
            k = q_match.start(0)
        if q_match and (not o_match or k < j):
            _push(b[i:k])
            _push(b[k + 1 : k + 2])
            i = k + 2
        else:
            _push(b[i:j])
            rv.append(int(b[j + 1 : j + 4], 8))
            i = j + 4

    return bytes(rv)


def _cookie_parse_impl(b):
    """Lowlevel cookie parsing facility that operates on bytes."""
    i = 0
    n = len(b)

    while i < n:
        match = _cookie_re.search(b + b";", i)
        if not match:
            break

        key = match.group("key").strip()
        value = match.group("val") or b""
        i = match.end(0)

        yield _cookie_unquote(key), _cookie_unquote(value)


def _encode_idna(domain):
    # If we're given bytes, make sure they fit into ASCII
    if not isinstance(domain, text_type):
        domain.decode("ascii")
        return domain

    # Otherwise check if it's already ascii, then return
    try:
        return domain.encode("ascii")
    except UnicodeError:
        pass

    # Otherwise encode each part separately
    parts = domain.split(".")
    for idx, part in enumerate(parts):
        parts[idx] = part.encode("idna")
    return b".".join(parts)


def _decode_idna(domain):
    # If the input is a string try to encode it to ascii to
    # do the idna decoding.  if that fails because of an
    # unicode error, then we already have a decoded idna domain
    if isinstance(domain, text_type):
        try:
            domain = domain.encode("ascii")
        except UnicodeError:
            return domain

    # Decode each part separately.  If a part fails, try to
    # decode it with ascii and silently ignore errors.  This makes
    # most sense because the idna codec does not have error handling
    parts = domain.split(b".")
    for idx, part in enumerate(parts):
        try:
            parts[idx] = part.decode("idna")
        except UnicodeError:
            parts[idx] = part.decode("ascii", "ignore")

    return ".".join(parts)


def _make_cookie_domain(domain):
    if domain is None:
        return None
    domain = _encode_idna(domain)
    if b":" in domain:
        domain = domain.split(b":", 1)[0]
    if b"." in domain:
        return domain
    raise ValueError(
        "Setting 'domain' for a cookie on a server running locally (ex: "
        "localhost) is not supported by complying browsers. You should "
        "have something like: '127.0.0.1 localhost dev.localhost' on "
        "your hosts file and then point your server to run on "
        "'dev.localhost' and also set 'domain' for 'dev.localhost'"
    )


def _easteregg(app=None):
    """Like the name says.  But who knows how it works?"""

    def bzzzzzzz(gyver):
        import base64
        import zlib

        return zlib.decompress(base64.b64decode(gyver)).decode("ascii")

    gyver = u"\n".join(
        [
            x + (77 - len(x)) * u" "
            for x in bzzzzzzz(
                b"""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"""
            ).splitlines()
        ]
    )

    def easteregged(environ, start_response):
        def injecting_start_response(status, headers, exc_info=None):
            headers.append(("X-Powered-By", "Werkzeug"))
            return start_response(status, headers, exc_info)

        if app is not None and environ.get("QUERY_STRING") != "macgybarchakku":
            return app(environ, injecting_start_response)
        injecting_start_response("200 OK", [("Content-Type", "text/html")])
        return [
            (
                u"""
<!DOCTYPE html>
<html>
<head>
<title>About Werkzeug</title>
<style type="text/css">
  body { font: 15px Georgia, serif; text-align: center; }
  a { color: #333; text-decoration: none; }
  h1 { font-size: 30px; margin: 20px 0 10px 0; }
  p { margin: 0 0 30px 0; }
  pre { font: 11px 'Consolas', 'Monaco', monospace; line-height: 0.95; }
</style>
</head>
<body>
<h1><a href="http://werkzeug.pocoo.org/">Werkzeug</a></h1>
<p>the Swiss Army knife of Python web development.</p>
<pre>%s\n\n\n</pre>
</body>
</html>"""
                % gyver
            ).encode("latin1")
        ]

    return easteregged



from mw_helper import ensure_iterable
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates


class KeepProdStateBox(OrderChangerProductionBox):
    def __init__(self, name, sons, states, logger=None):
        super(KeepProdStateBox, self).__init__(name, sons, logger)
        states = ensure_iterable(states)
        self.allowed_states = {}
        for state in states:
            self.allowed_states[state] = state

    def change_order(self, order):
        if order.prod_state in self.allowed_states:
            order.prod_state = ProdStates.NORMAL

        return order



from production.model.request import ChangePathRequest

from ._ChangePathProcessor import ChangePathProcessor


class DisablePathProcessor(ChangePathProcessor):
    def parse_data(self, data, event_type):
        return ChangePathRequest(data, False)



from abc import ABCMeta

from messagebus import (
    DefaultToken,
    InternalErrorMessage,
    Message,
    MessageBus,
)
from messageprocessorutil import UuidMessageProcessor
from typing import (
    Any,
    Optional,
)


class BaseProcessor(UuidMessageProcessor):
    __metaclass__ = ABCMeta

    def __init__(
        self,   # type: BaseProcessor
    ):
        # type: (...) -> None
        super(BaseProcessor, self).__init__()

    def format_response(
        self,           # type: BaseProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        request,        # type: Any
        result,         # type: Any
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=Message(token=DefaultToken.TK_SYS_ACK),
        )

    def format_exception(
        self,           # type: BaseProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        input_obj,      # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(ex=exception),
        )

    def format_parse_exception(
        self,           # type: BaseProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        data,           # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=data,
            exception=exception,
        )



from sysactions import action

from .common_lock import (
    acquire_user_lock,
    release_user_lock,
)
from .. import logger
from ..models import UserLockType


@action
def acquire_user_lock_performing_tender_action(
    pos_id,             # type: str
    user_id,            # type: str
):
    # type: (...) -> bool

    log_message = "Acquire user lock performing tender action POS: {} - User: {}"
    logger.debug(log_message.format(pos_id, pos_id))
    three_minutes = 60 * 3
    return acquire_user_lock(
        user_id=user_id,
        lock_reason=UserLockType.PERFORMING_TENDER,
        lock_timeout=three_minutes
    )


@action
def release_user_lock_action(
    pos_id,     # type: str
    user_id,    # type: str
):
    # type: (...) -> None

    logger.debug("Release user lock action POS: {} - User: {}".format(pos_id, user_id))
    release_user_lock(
        user_id=user_id,
    )


def acquire_user_lock_close_operator(
    user_id,    # type: str
):
    # type: (...) -> bool

    return acquire_user_lock(
        user_id=user_id,
        lock_reason=UserLockType.CLOSING_OPERATOR,
    )



from abc import ABCMeta, abstractmethod

from messagebus import Event
from typing import List


class Waiter(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_sync_id(self):
        # type: () -> str
        raise NotImplementedError()

    def get_events(self):
        # type: () -> List[Event]
        raise NotImplementedError()



import os

import msgbus
from systools import sys_parsefilepath

from application.model.customexception import NoUpdateFound
from typing import Optional
from threading import Thread
from application.interactor import CatalogUpdaterInteractor
from application.manager._UpdaterManager import UpdaterManager
from application.model import UpdateType, UpdateStep, UpdateStatus
from application.model.configuration import Configurations
from application.repository.api import CatalogUpdaterAPIRepository
from application.repository.pos import (
    BinaryUpdaterPOSRepository,
    CatalogUpdaterPOSRepository,
    CommonPOSRepository,
)
from application.service import (
    MediaUpdaterService,
    UpdaterService,
)
from application.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)
MAX_ATTEMPT = 3

GENESIS_PATH = "../../genesis"


class CatalogUpdaterManager(UpdaterManager):
    def __init__(
        self,
        configs,                    # type: Configurations
        interactor,                 # type: CatalogUpdaterInteractor
        api_repository,             # type: CatalogUpdaterAPIRepository
        pos_repository,             # type: CatalogUpdaterPOSRepository
        updater_service,            # type: UpdaterService
        media_updater_service,      # type: MediaUpdaterService,
        binaries_repository,        # type: BinaryUpdaterPOSRepository
        common_pos_repository,      # type: CommonPOSRepository
    ):
        # type: (...) -> None # noqa

        name = UpdateType.catalog.name
        super(CatalogUpdaterManager, self).__init__(name, configs, interactor, pos_repository, logger)
        self.configs = configs
        self.update_type = UpdateType.catalog
        self.local_configs = configs.updaters[self.update_type.name]
        self.active_frequency = int(self.local_configs.active_frequency)
        self.interactor = interactor
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.updater_service = updater_service
        self.media_updater_service = media_updater_service
        self.binaries_repository = binaries_repository
        self.common_pos_repository = common_pos_repository

        data_path = sys_parsefilepath("{HVDATADIR}")
        self.data_relative_path = "data"
        self.data_path = os.path.join(data_path, GENESIS_PATH, self.data_relative_path)
        self.genesis_databases_path = os.path.join(data_path, "../../genesis/data/server/databases")

    def run(self, validate_window=True):
        # type: (Optional[bool]) -> None

        self._verify_necessary_product_info()

        catalog_notify_thread = Thread(target=self._notify_update, name="Catalog Notify")
        catalog_notify_thread.daemon = True
        catalog_notify_thread.start()

        super(CatalogUpdaterManager, self).run()

    def _verify_necessary_product_info(self):
        product_db = os.path.join(self.genesis_databases_path, "product.db")
        if not os.path.exists(product_db):
            return

        table_exists = self.common_pos_repository.table_exists_in_db(
            database_path=product_db,
            table_name="PosSaleTypeNavigation",
        )
        if not table_exists or not self.common_pos_repository.has_pos_sale_type_navigation_populated(product_db):
            self.update(force=True)

    def has_update_to_apply(self):
        # type: () -> None

        if not self.binaries_repository.has_already_applied_update():
            raise NoUpdateFound()

    def update(
        self,               # type: CatalogUpdaterManager
        force=False,        # type: bool
    ):
        # type: (...) -> None

        self.updater_service.download_campaign_pending_version(force=force)
        self.interactor.apply_update(force=force)

    def _notify_update(self):
        # type: () -> None

        while self.running:
            try:
                database_id, update_id = self.pos_repository.get_pending_update_to_notify()
                if update_id:
                    update_id = int(update_id)
                    logger.info("Notifying update {}".format(str(update_id)))
                    self.api_repository.notify_applied_update(database_id, update_id)
                    self.pos_repository.update_step(update_id, UpdateStep.notify)
                    self.pos_repository.update_status(update_id, UpdateStatus.applied)
                    logger.info("Update {} successfully notified!".format(str(update_id)))
                    self._clean_rupture()
                    self.media_updater_service.update_media(force_update=True)
            except Exception as _:
                logger.exception("Error on automatically notify update")
            finally:
                database_id, update_id = self.pos_repository.get_pending_update_to_notify()
                if database_id is not None and update_id is not None:
                    self.sleep(30)
                else:
                    self.sleep(self.active_frequency)

    def _clean_rupture(self):
        # type: ()-> ()

        try:
            logger.info("Start clean rupture")
            attempt = 0
            while attempt < MAX_ATTEMPT:
                message = self.pos_repository.clean_rupture()
                if message.token == msgbus.TK_SYS_ACK:
                    break
                attempt += 1
                logger.info("Clean rupture attempt: {}".format(attempt))
                self.sleep(1)
        except Exception as _:
            logger.exception("Error on clean rupture")

        logger.info("End clean rupture")



from logging import Logger

from services.services import UpdateService


class BOHChangeNotificationConfigInteractor(object):
    def __init__(
        self,
        update_service,  # type: UpdateService
        logger,          # type: Logger
    ):
        # type: (...) -> None

        self.update_service = update_service
        self.logger = logger

    def process_request(
        self,
    ):
        # type: (...) -> None

        try:
            self.update_service.update_configured_alerts()
        except Exception:  # noqa
            self.logger.exception("Error updating configured alerts")
            raise
        else:
            self.update_service.update_configured_alerts_ack()





# -*- coding: utf-8 -*-

from bustoken import TK_REMOTE_ORDER_SAVE_ORDER
from msgbus import TK_SYS_ACK

from .. import mb_context
from model.customexception import SaveDeliveryOrderError


def save_delivery_order(order_id):
    data = order_id
    msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_SAVE_ORDER, data=data, timeout=10000000)
    if msg.token == TK_SYS_ACK:
        return
    
    raise SaveDeliveryOrderError("$ERROR_SAVING_ORDER")



from open_loyalty import (
    build_loyalty_properties,
    get_document_from_loyalty_or_customer_response,
    get_order_loyalty_customer_attribute,
    migrate_order_loyalty_info_to_table,
    persist_loyalty_properties_ts,
    parse_table_loyalty_customer_info,
    get_table_loyalty_customer_attribute,
)

from get_loyalty_document import get_loyalty_document



from string import upper

from sysactions import (
    action,
    get_model,
    get_poslist,
    get_podtype,
    get_custom,
    set_custom,
    show_listbox,
    translate_message,
)

from xml.etree import cElementTree as eTree

from typing import List


@action
def open_close_totem(
    pos_id,  # type: int
):
    # type: (...) -> None

    totem_id_list = _get_totem_list()
    options = _get_options(pos_id, totem_id_list)
    selected_index = show_listbox(
        posid=pos_id,
        options=options,
        message="$OPEN_CLOSE_TOTEM",
    )
    if selected_index is None:
        return

    index = int(selected_index)
    totem_info = [x for x in totem_id_list if x["index"] == index][0]
    set_custom(
        posid=totem_info["totem_id"],
        key='TOTEM_WAS_CLOSED',
        value=not totem_info["status"],
    )


def _get_options(
    pos_id,         # int
    totem_id_list,  # type: List
):
    # type: (...) -> List

    model = get_model(posid=pos_id)
    options = []
    status_translation_dict = _get_status_translation_dict(model=model)
    for c in totem_id_list:
        if c["status"]:
            status = "({})".format(status_translation_dict["CLOSED"])
        else:
            status = "({})".format(status_translation_dict["OPENED"])
        options.append(status_translation_dict["TOTEM_ID_STATUS"].format(c["totem_id"], status))

    return options


def _get_totem_list():
    # type: (...) -> List

    totem_id_list = []
    index_list = 0
    for pos_list_id in get_poslist():
        totem_info = _get_totem_info(
            index_list=index_list,
            pos_list_id=pos_list_id,
        )
        if totem_info is not None:
            totem_id_list.append(totem_info)
            index_list += 1

    return totem_id_list


def _get_status_translation_dict(
    model,  # type: eTree.XML
):
    # type: (...) -> dict

    status_translation_dict = {
        "CLOSED": translate_message(
            model=model,
            message_key="CLOSED",
        ),
        "OPENED": translate_message(
            model=model,
            message_key="OPENED",
        ),
        "TOTEM_ID_STATUS": translate_message(
            model=model,
            message_key="TOTEM_ID_STATUS",
        ),
    }
    return status_translation_dict


def _get_totem_info(
    index_list,  # type: int
    pos_list_id,  # type: int
):
    # type: (...) -> dict

    model = get_model(posid=pos_list_id)
    pod_type = get_podtype(model=model)
    if pod_type == "TT":
        totem_was_closed = get_custom(
            model=model,
            key='TOTEM_WAS_CLOSED',
        )
        status = False
        if totem_was_closed is None:
            status = False
        else:
            if upper(totem_was_closed) == "TRUE":
                status = True
            elif upper(totem_was_closed) == "FALSE":
                status = False
        return {
            "index": index_list,
            "status": status,
            "totem_id": pos_list_id,
        }



# epic.py - functions for handling Indian voter identification numbers
#
# Copyright (C) 2021 Gaurav Chauhan
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""EPIC (Electoral Photo Identity Card, Indian Voter ID).

The Electoral Photo Identity Card (EPIC) is an identity document issued by
the Election Commission of India (ECI) only to the India citizens who have
reached the age of 18.

Each EPIC contains an unique 10 digit alphanumeric identifier known as EPIC
number or Voter ID number.

Every EPIC number begins with a Functional Unique Serial Number (FUSN), a 3
letter unique identifier for each Assembly Constituency. FUSN is followed by
a 6 digit serial number and 1 check digit of the serial number calculated
using Luhn algorithm.

More information:

* https://en.wikipedia.org/wiki/Voter_ID_(India)
* https://www.kotaksecurities.com/ksweb/voter-id/serial-number-in-elctoral-roll

>>> validate('WKH1186253')
'WKH1186253'
>>> validate('WKH118624')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('1231186253')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('WKH1186263')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

import re

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean


_EPIC_RE = re.compile(r'^[A-Z]{3}[0-9]{7}$')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


def validate(number):
    """Check if the number provided is a valid EPIC number. This checks the
    length, formatting and checksum."""
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength
    if not _EPIC_RE.match(number):
        raise InvalidFormat()
    luhn.validate(number[3:])
    return number


def is_valid(number):
    """Check if the number provided is a valid EPIC number. This checks the
    length, formatting and checksum."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False
